#!/usr/bin/env python

from . import plugins, testmodel
import os
import sys
import signal
import operator
import logging
from threading import Thread
from collections import OrderedDict
from time import sleep
from glob import glob
from copy import copy
import time
from functools import reduce, cmp_to_key

# Class to allocate unique names to tests for script identification and cross process communication


class UniqueNameFinder(plugins.Responder):
    def __init__(self, optionMap, *args):
        plugins.Responder.__init__(self, optionMap)
        self.name2test = {}
        self.diag = logging.getLogger("Unique Names")

    def notifyAdd(self, test, *args, **kw):
        if test.name in self.name2test:
            oldTest = self.name2test[test.name]
            self.storeUnique(oldTest, test)
        else:
            self.storeBothWays(test.name, test)

    def notifyRemove(self, test):
        self.removeName(test.name)

    def removeName(self, name):
        if name in self.name2test:
            self.diag.info("Removing test name " + name)
            del self.name2test[name]

    def notifyNameChange(self, test, origRelPath):
        oldName = os.path.basename(origRelPath)
        self.removeName(oldName)
        self.notifyAdd(test)

    def findParentIdentifiers(self, oldTest, newTest):
        oldParentId = " at top level"
        if oldTest.parent:
            oldParentId = " under " + oldTest.parent.name
        newParentId = " at top level"
        if newTest.parent:
            newParentId = " under " + newTest.parent.name
        if oldTest.parent and newTest.parent and oldParentId == newParentId:
            oldNextLevel, newNextLevel = self.findParentIdentifiers(oldTest.parent, newTest.parent)
            oldParentId += oldNextLevel
            newParentId += newNextLevel
        return oldParentId, newParentId

    def storeUnique(self, oldTest, newTest):
        oldParentId, newParentId = self.findParentIdentifiers(oldTest, newTest)
        if oldParentId != newParentId:
            self.storeBothWays(oldTest.name + oldParentId, oldTest)
            self.storeBothWays(newTest.name + newParentId, newTest)
        elif oldTest.app.name != newTest.app.name:
            self.storeBothWays(oldTest.name + " for " + oldTest.app.fullName(), oldTest)
            self.storeBothWays(newTest.name + " for " + newTest.app.fullName(), newTest)
        elif oldTest.app.getFullVersion() != newTest.app.getFullVersion():
            self.storeBothWays(oldTest.name + " version " + self.getVersionName(oldTest), oldTest)
            self.storeBothWays(newTest.name + " version " + self.getVersionName(newTest), newTest)
        else:
            # Overwrite the old with the new if they can't be distinguished
            self.storeBothWays(newTest.name + newParentId, newTest)

    def getVersionName(self, test):
        version = test.app.getFullVersion()
        if len(version):
            return version
        else:
            return "<default>"

    def storeBothWays(self, name, test):
        self.diag.info("Setting unique name for test " + test.name + " to " + name)
        self.name2test[name] = test
        test.setUniqueName(name)


class Activator(plugins.Responder, plugins.Observable):
    def __init__(self, optionMap, allApps):
        plugins.Responder.__init__(self, optionMap, allApps)
        plugins.Observable.__init__(self)
        self.allowEmpty = "gx" in optionMap or optionMap.runScript()
        self.suites = []
        self.diag = logging.getLogger("Activator")
        self.suiteCopyCache = {}

    def addSuites(self, suites):
        self.suites = suites

    def findGuideSuiteForCopy(self, versions):
        versionsForGuide = tuple([v for v in versions if "copy_" not in v])
        if versionsForGuide in self.suiteCopyCache:
            return self.suiteCopyCache.get(versionsForGuide)

        for suite in self.suites:
            if tuple(suite.app.versions) == versionsForGuide:
                self.suiteCopyCache[versionsForGuide] = suite
                return suite

    def readTestSuiteContents(self, suite):
        if any(("copy_" in v) for v in suite.app.versions):
            guideSuite = self.findGuideSuiteForCopy(suite.app.versions)
            if guideSuite is not None:
                self.diag.info("Creating test suite by copying " + repr(guideSuite))
                return suite.readContents(guideSuite=guideSuite)

        filters = suite.app.getFilterList(self.suites)
        self.diag.info("Creating test suite with filters " + repr(filters))
        return suite.readContents(filters)

    def run(self):
        goodSuites = []
        rejectionInfo = OrderedDict()
        self.notify("StartRead")
        for suite in self.suites:
            try:
                self.readTestSuiteContents(suite)
                self.diag.info("SUCCESS: Created test suite of size " + str(suite.size()))

                if suite.size() > 0 or self.allowEmpty:
                    goodSuites.append(suite)
                    suite.notify("Add", initial=True)
                else:
                    rejectionInfo[suite.app] = "no tests matching the selection criteria found."
            except plugins.TextTestError as e:
                rejectionInfo[suite.app] = str(e)

        self.notify("AllRead", goodSuites)

        if len(rejectionInfo) > 0:
            self.writeErrors(rejectionInfo)
        # triggers the ActionRunner to start if needed, do this in the same thread!
        self.performNotify("AllReadAndNotified")
        return goodSuites

    def writeErrors(self, rejectionInfo):
        # Don't write errors if only some of a group are rejected
        appsByName = OrderedDict()
        rejectedApps = set(rejectionInfo.keys())
        for suite in self.suites:
            app = suite.app
            appsByName.setdefault(app.name, []).append(app)

        for _, appGroup in list(appsByName.items()):
            if set(appGroup).issubset(rejectedApps):
                for app in appGroup:
                    if app in rejectionInfo:
                        sys.stderr.write(app.rejectionMessage(rejectionInfo.get(app)))


class TextTest(plugins.Responder, plugins.Observable):
    def __init__(self):
        plugins.Responder.__init__(self)
        plugins.Observable.__init__(self)
        if os.name == "posix":
            # To aid in debugging tests that hang...
            signal.signal(signal.SIGQUIT, self.printStackTrace)
        self.setSignalHandlers(self.handleSignalWhileStarting)
        self.inputOptions = testmodel.OptionFinder()
        self.diag = logging.getLogger("Find Applications")
        self.appSuites = OrderedDict()
        self.exitCode = 0

    def printStackTrace(self, *args):
        # Swiped from http://stackoverflow.com/questions/1032813/dump-stacktraces-of-all-active-threads
        # print_stack() only shows the main thread which is often just waiting for the others.
        sys.stderr.write("Received SIGQUIT: showing current stack trace below:\n")
        code = []
        from traceback import extract_stack
        for threadId, stack in list(sys._current_frames().items()):
            code.append("# ThreadID: %s" % threadId)
            for filename, lineno, name, line in extract_stack(stack):
                code.append('  File "%s", line %d, in %s' % (filename, lineno, name))
                if line:
                    code.append("    " + line.strip())

        for line in code:
            sys.stderr.write(line + "\n")

    def findSearchDirs(self, roots):
        subDirs = []
        for root in roots:
            for f in sorted(os.listdir(root)):
                path = os.path.join(root, f)
                if os.path.isdir(path):
                    subDirs.append(path)
        return roots + subDirs

    def findApps(self):
        roots = list(filter(os.path.isdir, self.inputOptions.rootDirectories))
        if len(roots) == 0:
            for root in self.inputOptions.rootDirectories:
                sys.stderr.write("Test suite root directory does not exist: " + root + "\n")
            return True, []

        # Just ignore any roots that don't exist
        self.inputOptions.rootDirectories = roots
        self.diag.info("Using test suite at " + repr(roots))
        searchDirs = self.findSearchDirs(roots)
        if "new" in self.inputOptions:
            return False, []

        appList = []
        raisedError = False
        selectedAppDict = self.inputOptions.findSelectedAppNames()
        for dir in searchDirs:
            ignoreNames = [app.name for app in appList]
            subRaisedError, apps = self.findAppsUnder(dir, selectedAppDict, ignoreNames)
            appList += apps
            raisedError |= subRaisedError

        if not raisedError:
            for missingAppName in self.findMissingApps(appList, list(selectedAppDict.keys())):
                sys.stderr.write("Could not read application '" + missingAppName + "'. No file named config." +
                                 missingAppName + " was found under " + " or ".join(self.inputOptions.rootDirectories) + ".\n")
                raisedError = True

        appList.sort(key=lambda app: app.name)
        self.diag.info("Found applications : " + repr(appList))
        return raisedError, appList

    def findMissingApps(self, appList, selectedApps):
        return [appName for appName in selectedApps if self.appMissing(appName, appList)]

    def appMissing(self, appName, apps):
        return reduce(operator.and_, (app.name != appName for app in apps), True)

    def findAppsUnder(self, dirName, selectedAppDict, ignoreNames):
        appList = []
        raisedError = False
        self.diag.info("Selecting apps in " + dirName + " according to dictionary :" + repr(selectedAppDict))
        dircache = testmodel.DirectoryCache(dirName)
        for f in dircache.findAllFiles("config"):
            if not os.path.isfile(f):
                continue  # ignore broken links and directories
            components = os.path.basename(f).split('.')
            if len(components) != 2:
                continue
            appName = components[1]

            # Ignore emacs backup files and stuff we haven't selected
            if appName.endswith("~") or (len(selectedAppDict) and appName not in selectedAppDict) or appName in ignoreNames:
                continue

            self.diag.info("Building apps from " + f)
            versionList = self.inputOptions.findVersionList()
            if appName in selectedAppDict:
                versionList = selectedAppDict[appName]
            extraVersionsDuplicating = []
            for versionStr in versionList:
                appVersions = list(filter(len, versionStr.split(".")))  # remove empty versions
                app, currExtra = self.addApplication(appName, dircache, appVersions, versionList)
                if app:
                    appList.append(app)
                    extraVersionsDuplicating += currExtra
                else:
                    raisedError = True
            for toRemove in [app for app in appList if app.getFullVersion() in extraVersionsDuplicating]:
                appList.remove(toRemove)
        return raisedError, appList

    def createApplication(self, appName, dircache, versions):
        try:
            return testmodel.Application(appName, dircache, versions, self.inputOptions)
        except (testmodel.BadConfigError, plugins.TextTestError) as e:
            sys.stderr.write("Unable to load application from file 'config." + appName + "' - " + str(e) + ".\n")

    def addApplication(self, appName, dircache, appVersions, allVersions=[]):
        app = self.createApplication(appName, dircache, appVersions)
        if not app:
            return None, []
        extraVersionsDuplicating = []
        for extraVersion in app.getExtraVersions():
            if extraVersion in allVersions:
                extraVersionsDuplicating.append(extraVersion)
            if "copy_" in extraVersion:
                extraApp = copy(app)
                extraApp.versions = appVersions + extraVersion.split(".")
                extraApp.extras = []
                extraApp.checkout = extraApp.configObject.setUpCheckout(extraApp)
            else:
                extraApp = self.createApplication(appName, dircache, appVersions + extraVersion.split("."))
            if extraApp:
                # Autogenerated extra versions are marked unsaveable on the parent,
                # obviously needs transferring to the extraApp here
                if extraVersion in app.getConfigValue("unsaveable_version"):
                    extraApp.addConfigEntry("unsaveable_version", extraVersion)
                app.extras.append(extraApp)
        return app, extraVersionsDuplicating

    def getAllConfigObjects(self, allApps):
        if len(allApps) > 0:
            return allApps
        else:
            return [plugins.importAndCall("default", "getConfig", self.inputOptions)]

    def createResponders(self, allApps):
        responderClasses = self.getBuiltinResponderClasses()
        for configObject in self.getAllConfigObjects(allApps):
            for respClass in configObject.getResponderClasses(allApps):
                if not respClass in responderClasses:
                    self.diag.info("Adding responder " + repr(respClass))
                    responderClasses.insert(-2, respClass)  # keep Activator and AllCompleteResponder at the end
        self.removeBaseClasses(responderClasses)
        self.diag.info("Filtering away base classes, using " + repr(responderClasses))
        self.observers = [x(self.inputOptions, allApps) for x in responderClasses]

    def getBuiltinResponderClasses(self):
        return [UniqueNameFinder, Activator, testmodel.AllCompleteResponder]

    def removeBaseClasses(self, classes):
        # Different apps can produce different versions of the same responder/thread runner
        # We should make sure we only include the most specific ones, in the furthest forward position their relatives achieved...
        # (the order can be important)
        toRemove = []
        newPositions = {}
        for i, class1 in enumerate(classes):
            for class2 in classes[i+1:]:
                if issubclass(class1, class2):
                    toRemove.append(class2)
                elif issubclass(class2, class1):
                    toRemove.append(class1)
                    if class2 not in newPositions:
                        newPositions[class2] = i

        for cls, i in list(newPositions.items()):
            classes.remove(cls)
            classes.insert(i, cls)
        for cls in toRemove:
            classes.remove(cls)

    def createTestSuites(self, allApps):
        appSuites = OrderedDict()
        raisedError = False
        for app in allApps:
            warningMessages = []
            appGroup = [app] + app.extras
            for partApp in appGroup:
                try:
                    testSuite = self.createInitialTestSuite(partApp)
                    appSuites[partApp] = testSuite
                except plugins.TextTestWarning as e:
                    warningMessages.append(partApp.rejectionMessage(str(e)))
                except plugins.TextTestError as e:
                    sys.stderr.write(partApp.rejectionMessage(str(e)))
                    raisedError = True
                except Exception:
                    sys.stderr.write("Error creating test suite for " + partApp.description() + " :\n")
                    plugins.printException()
            fullMsg = "".join(warningMessages)
            # If the whole group failed, we write to standard error, where the GUI will find it. Otherwise we just log in case anyone cares.
            if len(warningMessages) == len(appGroup):
                sys.stderr.write(fullMsg)
                raisedError = True
            else:
                sys.stdout.write(fullMsg)
        return raisedError, appSuites

    def notifyExit(self):
        # Can get called several times, protect against this...
        if len(self.appSuites) > 0:
            self.notify("Status", "Removing all temporary files ...")
            for app, testSuite in self.appSuites.items():
                self.notify("ActionProgress")
                app.cleanWriteDirectory(testSuite)
            self.notify("Status", "Removed all temporary files ...")
            self.appSuites = []

    def notifyComplete(self, test):
        testExitCode = test.state.getExitCode()
        if testExitCode and self.exitCode != 1:
            self.exitCode = testExitCode

    def run(self):
        try:
            self._run()
            self.diag.info("Exiting with exit code " + str(self.exitCode))
            sys.exit(self.exitCode)
        except plugins.TextTestError as e:
            sys.stderr.write(str(e) + "\n")
            sys.exit(1)
        except KeyboardInterrupt:
            pass  # already written about this

    def _run(self):
        appFindingWroteError, allApps = self.findApps()
        if self.inputOptions.helpMode():
            if len(allApps) > 0:
                allApps[0].printHelpText()
            else:
                print("TextTest didn't find any valid test applications - you probably need to tell it where to find them.")
                print("The most common way to do this is to set the environment variable TEXTTEST_HOME.")
                print("If this makes no sense, read the online documentation...")
                print(testmodel.helpIntro)
            return

        if len(allApps) == 0 and appFindingWroteError:
            self.exitCode = 1
            return

        if self.inputOptionsValid(allApps):
            try:
                self.createAndRunSuites(allApps)
            finally:
                self.notifyExit()  # include the dud ones, possibly

    def inputOptionsValid(self, allApps):
        validOptions = self.findAllValidOptions(allApps)
        for option in list(self.inputOptions.keys()):
            if option not in validOptions:
                sys.stderr.write("texttest: unrecognised option '-" + option + "'\n")
                self.exitCode = 1
                return False
        return True

    def findAllValidOptions(self, allApps):
        validOptions = set()
        for configObject in self.getAllConfigObjects(allApps):
            validOptions.update(set(configObject.findAllValidOptions(allApps)))
        return validOptions

    def createAndRunSuites(self, allApps):
        self.createResponders(allApps)
        raisedError, self.appSuites = self.createTestSuites(allApps)
        if not raisedError or len(self.appSuites) > 0:
            self.addSuites(list(self.appSuites.values()), allApps)

            # Set the signal handlers to use when running, if we actually plan to do any
            self.setSignalHandlers(self.handleSignal)

            self.runThreads()
        if raisedError:
            self.exitCode = 1

    def addSuites(self, emptySuites, allApps):
        for object in self.observers:
            # For all observable responders, set them to be observed by the others if they
            # haven't fixed their own observers
            if isinstance(object, plugins.Observable) and len(object.observers) == 0:
                self.diag.info("All responders now observing " + str(object.__class__))
                object.setObservers([self] + self.observers)
            suites = self.getSuitesToAdd(object, emptySuites, allApps)
            self.diag.info("Adding suites " + repr(suites) + " for " + str(object.__class__))
            object.addSuites(suites)

    def getSuitesToAdd(self, observer, emptySuites, allApps):
        for responderClass in self.getBuiltinResponderClasses():
            if isinstance(observer, responderClass):
                return emptySuites

        responderClassDict = {}
        for app in allApps:
            responderClassDict[app.name] = app.getResponderClasses(list(self.appSuites.keys()))

        suites = []
        for testSuite in emptySuites:
            for responderClass in responderClassDict.get(testSuite.app.name):
                if isinstance(observer, responderClass):
                    suites.append(testSuite)
                    break
        return suites

    def getRootSuite(self, appName, versions):
        for app, testSuite in list(self.appSuites.items()):
            if app.name == appName and app.versions == versions:
                return testSuite

        dirCache = self.makeDirectoryCache(appName)
        if dirCache:
            newApp = self.addApplication(appName, dirCache, versions)[0]
            return self.createEmptySuite(newApp)
        else:
            message = "Couldn't create directory cache for application: " + str(appName)
            plugins.printWarning(message)
            return None

    def createEmptySuite(self, newApp):
        emptySuite = self.createInitialTestSuite(newApp)
        self.appSuites[newApp] = emptySuite
        self.addSuites([emptySuite], [newApp])
        return emptySuite

    def createInitialTestSuite(self, app):
        return app.createInitialTestSuite([self] + self.observers)

    def makeDirectoryCache(self, appName):
        configFile = "config." + appName
        for rootDir in self.inputOptions.rootDirectories:
            rootConfig = os.path.join(rootDir, configFile)
            if os.path.isfile(rootConfig):
                return testmodel.DirectoryCache(rootDir)
            else:
                allFiles = glob(os.path.join(rootDir, "*", configFile))
                if len(allFiles) > 0:
                    return testmodel.DirectoryCache(os.path.dirname(allFiles[0]))

    def notifyExtraTest(self, testPath, appName, versions):
        rootSuite = self.getRootSuite(appName, versions)
        if rootSuite:
            rootSuite.addTestCaseWithPath(testPath)
        else:
            message = "Couldn't add extra test for application: " + str(appName)
            plugins.printWarning(message)

    def notifyNewApplication(self, newApp):
        suite = self.createEmptySuite(newApp)
        suite.notify("Add", initial=False)

    def findThreadRunners(self):
        allRunners = [x for x in self.observers if hasattr(x, "run")]
        mainThreadRunner = None
        for runner in allRunners:
            if runner.canBeMainThread():
                mainThreadRunner = runner
                break
        allRunners.remove(mainThreadRunner)
        return mainThreadRunner, allRunners

    def runThreads(self):
        # Run the first one as the main thread and the rest in subthreads
        # Make sure all of them are finished before we stop
        mainThreadRunner, subThreadRunners = self.findThreadRunners()
        allThreads = []
        for subThreadRunner in subThreadRunners:
            thread = Thread(target=subThreadRunner.run, name=subThreadRunner.__class__.__name__)
            allThreads.append(thread)
            self.diag.info("Running " + str(subThreadRunner.__class__) + " in a subthread")
            thread.start()

        if mainThreadRunner:
            self.diag.info("Running " + str(mainThreadRunner.__class__) + " in main thread")
            mainThreadRunner.run()

        self.waitForThreads(allThreads)

    def waitForThreads(self, allThreads):
        # Need to wait for the threads to terminate in a way that allows signals to be
        # caught. thread.join doesn't do this. signal.pause seems like a good idea but
        # doesn't return unless a signal is caught, leading to sending "fake" ones from the
        # threads when they finish. And playing with signals and threads together is playing with fire...

        # So we poll, which we don't really want to do, but it seems better than using Twisted or asyncore
        # just for this :) With a long enough sleep it shouldn't generate too much load...

        # See http://groups.google.com/group/comp.lang.python/browse_thread/thread/a244905b86f06e48/7e969a0c7932fa91#
        currThreads = self.aliveThreads(allThreads)
        threadCount = len(currThreads)
        while threadCount > 0:
            sleep(0.5)
            currThreads = self.aliveThreads(currThreads)
            if len(currThreads) < threadCount:
                self.diag.info("Thread(s) terminated, remaining are " + repr([t.name for t in currThreads]))
            threadCount = len(currThreads)

    def aliveThreads(self, threads):
        return [thread for thread in threads if thread.is_alive()]

    def getSignals(self):
        if hasattr(signal, "SIGUSR1"):
            # Signals used on UNIX to signify running out of CPU time, wallclock time etc.
            return [signal.SIGINT, signal.SIGTERM, signal.SIGUSR1, signal.SIGUSR2, signal.SIGXCPU]
        else:
            # Windows, which doesn't do signals
            return []

    def setSignalHandlers(self, handler):
        if os.name == "nt":
            try:
                import ctypes
                # Must store this, otherwise it gets garbage collected, and a crash results...
                self.wrapper = ctypes.WINFUNCTYPE(ctypes.c_int, ctypes.c_uint)(handler)
                # Without this, it will just ignore ctrl-C anyhow
                ctypes.windll.kernel32.SetConsoleCtrlHandler(None, False)
                ctypes.windll.kernel32.SetConsoleCtrlHandler(self.wrapper, True)
            except ImportError:
                pass # probably just documentation
        else:
            for sig in self.getSignals():
                signal.signal(sig, handler)

    def handleSignal(self, sig, *args):
        # Respond to the same signal only once and ignore all others!
        for ignoreSignal in self.getSignals():
            signal.signal(ignoreSignal, signal.SIG_IGN)
        signalText = self.getSignalText(sig)
        self.writeTermMessage(signalText)
        self.notify("Quit", sig)
        if len(self.appSuites) > 0:  # If the above succeeds in quitting they will be reset
            self.notify("KillProcesses", sig)
        if os.name == "nt":
            time.sleep(20)  # Time to clean up, process is killed when we exit
        else:
            return signalText

    def handleSignalWhileStarting(self, sig, *args):
        signalText = self.handleSignal(sig)
        if os.name != "nt":
            raise KeyboardInterrupt(signalText)

    def writeTermMessage(self, signalText):
        message = "Terminating testing due to external interruption"
        if signalText:
            message += " (" + signalText + ")"
        print(message)
        sys.stdout.flush()  # Try not to lose log file information...

    def getSignalText(self, sig):
        if hasattr(signal, "SIGUSR1"):
            if sig == signal.SIGUSR1:
                return "RUNLIMIT1"
            elif sig == signal.SIGXCPU:
                return "CPULIMIT"
            elif sig == signal.SIGUSR2:
                return "RUNLIMIT2"
        return ""  # mostly for historical reasons to be compatible with the default handler



import os
import string
import shutil
import sys
import logging
import glob
import re
from texttestlib import plugins
from configparser import ConfigParser, NoOptionError
from copy import copy
from collections import OrderedDict

plugins.addCategory("bug", "known bugs", "had known bugs")
plugins.addCategory("badPredict", "internal errors", "had internal errors")
plugins.addCategory("crash", "CRASHED")

# For backwards compatibility...


class FailedPrediction(plugins.TestState):
    def getExitCode(self):
        return int(self.category != "bug")

    def getTypeBreakdown(self):
        status = "failure" if self.getExitCode() else "success"
        return status, self.briefText


class Bug:
    rerunLine = "(NOTE: Test was run %d times in total and each time encountered this issue."
    prevResultLine = "Results of previous runs can be found in framework_tmp/backup.previous.* under the sandbox directory.)"

    def __init__(self, priority, rerunCount, rerunOnly):
        self.priority = priority
        self.rerunCount = rerunCount
        self.rerunOnly = rerunOnly

    def findCategory(self, internalError):
        if internalError or self.rerunCount:
            return "badPredict"
        else:
            return "bug"

    def isCancellation(self):
        return False

    def getRerunText(self):
        if self.rerunCount:
            return "\n" + self.rerunLine % (self.rerunCount + 1) + "\n" + self.prevResultLine + "\n\n"
        else:
            return ""


class BugSystemBug(Bug):
    def __init__(self, bugSystem, bugId, priorityStr, *args):
        self.bugId = bugId
        self.bugSystem = bugSystem
        prio = int(priorityStr) if priorityStr else 20
        Bug.__init__(self, prio, *args)

    def __repr__(self):
        return self.bugId

    def findInfo(self, test):
        location = test.getCompositeConfigValue("bug_system_location", self.bugSystem)
        username = test.getCompositeConfigValue("bug_system_username", self.bugSystem)
        password = test.getCompositeConfigValue("bug_system_password", self.bugSystem)
        status, bugText, isResolved, bugId = self.findBugInfo(self.bugId, location, username, password)
        self.bugId = bugId
        category = self.findCategory(isResolved)
        briefText = "bug " + self.bugId + " (" + status + ")"
        return category, briefText, self.getRerunText() + bugText

    def findBugInfo(self, bugId, location, username, password):
        namespace = {}
        exec("from ." + self.bugSystem + " import findBugInfo as _findBugInfo", globals(), namespace)
        return namespace["_findBugInfo"](self.bugId, location, username, password)  # @UndefinedVariable


class UnreportedBug(Bug):
    def __init__(self, fullText, briefText, internalError, priorityStr, *args):
        self.fullText = fullText
        self.briefText = briefText
        self.internalError = internalError
        prio = self.getPriority(priorityStr)
        Bug.__init__(self, prio, *args)

    def __repr__(self):
        return self.briefText

    def isCancellation(self):
        return not self.rerunOnly and not self.briefText and not self.fullText

    def getPriority(self, priorityStr):
        if priorityStr:
            return int(priorityStr)
        elif self.internalError:
            return 10
        else:
            return 30

    def findInfo(self, *args):
        return self.findCategory(self.internalError), self.briefText, self.getRerunText() + self.fullText


class BugTrigger:
    def __init__(self, getOption):
        useRegexp = int(getOption("use_regexp", "1"))
        searchStr = getOption("search_string").replace("\\n", "\n")
        self.textTrigger = plugins.MultilineTextTrigger(searchStr, useRegexp)
        self.triggerHosts = self.getTriggerHosts(getOption)
        self.checkUnchanged = int(getOption("trigger_on_success", "0"))
        self.reportInternalError = int(getOption("internal_error", "0"))
        self.ignoreOtherErrors = int(getOption("ignore_other_errors", self.reportInternalError))
        self.customTrigger = getOption("custom_trigger", "")
        self.bugInfo = self.createBugInfo(getOption)
        self.diag = logging.getLogger("Check For Bugs")

    def __repr__(self):
        return repr(self.textTrigger)

    def getTriggerHosts(self, getOption):
        hostStr = getOption("execution_hosts")
        if hostStr:
            return hostStr.split(",")
        else:
            return []

    def createBugInfo(self, getOption):
        bugSystem = getOption("bug_system")
        prioStr = getOption("priority")
        rerunCount = int(getOption("rerun_count", "0"))
        rerunOnly = int(getOption("rerun_only", "0"))
        if bugSystem:
            return BugSystemBug(bugSystem, getOption("bug_id"), prioStr, rerunCount, rerunOnly)
        else:
            return UnreportedBug(getOption("full_description"), getOption("brief_description"), self.reportInternalError, prioStr, rerunCount, rerunOnly)

    def matchesText(self, line):
        return self.textTrigger.matches(line)

    def exactMatch(self, lines, **kw):
        updatedLines = [line for i, line in enumerate(lines) if i < len(lines) - 1] if lines[-1] == '' else lines
        if len(updatedLines) == len(self.textTrigger.triggers):
            for index, line in enumerate(updatedLines, start=1):
                # We must check that every line match because MultilineTextTrigger.matches method
                # returns True only when the match is complete
                if index < len(updatedLines):
                    if not self.textTrigger._matches(line)[1]:
                        return False
                else:
                    return self.hasBug(line, **kw)
        return False

    def customTriggerMatches(self, *args):
        module, method = self.customTrigger.split(".", 1)
        return plugins.importAndCall(module, method, *args)

    def hasBug(self, line, execHosts=[], isChanged=True, multipleDiffs=False, tmpDir=None):
        if not self.checkUnchanged and not isChanged:
            self.diag.info("File not changed, ignoring")
            return False
        if multipleDiffs and not self.ignoreOtherErrors:
            self.diag.info("Multiple differences present, allowing others through")
            return False
        if line is not None and not self.textTrigger.matches(line):
            return False

        if self.customTrigger and not self.customTriggerMatches(execHosts, tmpDir):
            return False

        if self.hostsMatch(execHosts):
            return True
        else:
            self.diag.info("No match " + repr(execHosts) + " with " + repr(self.triggerHosts))
            return False

    def findBugInfo(self, test, fileStem, absenceBug):
        category, briefText, fullText = self.bugInfo.findInfo(test)
        whatText = "FAILING to find text" if absenceBug else "text found"
        matchText = repr(self)
        if "\n" in matchText:
            matchText = "'''\n" + matchText + "\n'''"
        else:
            matchText = "'" + matchText + "'"
        fullText += "\n(This bug was triggered by " + whatText + " in " + \
            self.getFileText(fileStem) + " matching " + matchText + ")"
        return category, briefText, fullText

    def getFileText(self, fileStem):
        if fileStem == "free_text":
            return "the full difference report"
        elif fileStem == "brief_text":
            return "the brief text/details"
        else:
            return "file " + repr(fileStem)

    def hostsMatch(self, execHosts):
        if len(self.triggerHosts) == 0:
            return True
        for host in execHosts:
            if not host in self.triggerHosts:
                return False
        return True


class FileBugData:
    def __init__(self):
        self.presentList = []
        self.absentList = []
        self.identicalList = []
        self.checkUnchanged = False
        self.diag = logging.getLogger("Check For Bugs")

    def addBugTrigger(self, getOption):
        bugTrigger = BugTrigger(getOption)
        if bugTrigger.checkUnchanged:
            self.checkUnchanged = True
        if getOption("trigger_on_absence", False):
            self.absentList.append(bugTrigger)
        elif getOption("trigger_on_identical", False):
            self.identicalList.append(bugTrigger)
        else:
            self.presentList.append(bugTrigger)

    def findBugs(self, fileName, execHosts, isChanged, multipleDiffs):
        if not self.checkUnchanged and not isChanged:
            self.diag.info("File not changed, ignoring all bugs")
            return []
        if not fileName:
            self.diag.info("File doesn't exist, checking only for absence bugs")
            return self.findAbsenceBugs(self.absentList, execHosts=execHosts, isChanged=isChanged, multipleDiffs=multipleDiffs, tmpDir=None)
        if not os.path.exists(fileName):
            raise plugins.TextTestError("The file '" + fileName +
                                        "' does not exist. Maybe it has been removed by an external process. ")

        self.diag.info("Looking for bugs in " + fileName)
        dirname = os.path.dirname(fileName)
        return self.findBugsInText(open(fileName).readlines(), execHosts=execHosts, isChanged=isChanged, multipleDiffs=multipleDiffs, tmpDir=dirname)

    def findBugsInText(self, lines, **kw):
        currAbsent = copy(self.absentList)
        bugs = []
        for bugTrigger in self.identicalList:
            if bugTrigger not in bugs and bugTrigger.exactMatch(lines, **kw):
                bugs.append(bugTrigger)
        for line in lines:
            self.diag.info("Checking " + repr(line))
            for bugTrigger in self.presentList:
                self.diag.info("Checking for existence of " + repr(bugTrigger))
                if bugTrigger not in bugs and bugTrigger.hasBug(line, **kw):
                    self.diag.info("FOUND!")
                    bugs.append(bugTrigger)
            toRemove = []
            for bugTrigger in currAbsent:
                self.diag.info("Checking for absence of " + repr(bugTrigger))
                if bugTrigger.matchesText(line):
                    self.diag.info("PRESENT!")
                    toRemove.append(bugTrigger)
            for bugTrigger in toRemove:
                currAbsent.remove(bugTrigger)

        return bugs + self.findAbsenceBugs(currAbsent, **kw)

    def findAbsenceBugs(self, absentList, **kw):
        bugs = []
        for bugTrigger in absentList:
            if bugTrigger not in bugs and bugTrigger.hasBug(None, **kw):
                bugs.append(bugTrigger)
        return bugs


class ParseMethod:
    def __init__(self, parser, section):
        self.parser = parser
        self.section = section

    def __call__(self, option, default=""):
        try:
            return self.parser.get(self.section, option)
        except NoOptionError:
            return default


class BugMap(OrderedDict):
    def checkUnchanged(self):
        for bugData in list(self.values()):
            if bugData.checkUnchanged:
                return True
        return False

    def readFromFile(self, fileName):
        parser = self.makeParser(fileName)
        if parser:
            self.readFromParser(parser)

    def readFromFileObject(self, f):
        parser = self.makeParserFromFileObject(f)
        if parser:
            self.readFromParser(parser)

    def makeParserFromFileObject(self, f):
        parser = ConfigParser()
        # Default behaviour transforms to lower case: we want case-sensitive
        parser.optionxform = str
        parser.readfp(f)
        return parser

    @staticmethod
    def makeParser(fileName):
        parser = ConfigParser()
        # Default behaviour transforms to lower case: we want case-sensitive
        parser.optionxform = lambda option: option
        try:
            parser.read(fileName)
            return parser
        except Exception as e:
            plugins.printWarning("Bug file at " + fileName + " could not be parsed, ignoring\n" + str(e))

    def readFromParser(self, parser):
        for section in reversed(sorted(parser.sections())):
            getOption = ParseMethod(parser, section)
            fileStem = getOption("search_file")
            self.setdefault(fileStem, FileBugData()).addBugTrigger(getOption)


class CheckForCrashes(plugins.Action):
    def __init__(self):
        self.diag = logging.getLogger("check for crashes")

    def __call__(self, test):
        if test.state.category == "killed":
            return
        # Hard-coded prediction: check test didn't crash
        comparison, _ = test.state.findComparison("stacktrace")
        if comparison and comparison.newResult():
            stackTraceFile = comparison.tmpFile
            self.diag.info("Parsing " + stackTraceFile)
            summary, errorInfo = self.parseStackTrace(test, stackTraceFile)

            newState = copy(test.state)
            newState.removeComparison("stacktrace")

            crashState = FailedPrediction("crash", errorInfo, summary)
            newState.setFailedPrediction(crashState)
            test.changeState(newState)
            if not test.app.keepTemporaryDirectories():
                os.remove(stackTraceFile)

    def parseStackTrace(self, test, stackTraceFile):
        lines = open(stackTraceFile).readlines()
        if len(lines) > 2:
            return lines[0].strip(), "".join(lines[2:])
        else:
            errFile = test.makeTmpFileName("stacktrace.collate_errs", forFramework=1)
            script = test.getCompositeConfigValue("collate_script", "stacktrace")[0]
            return "core not parsed", "The core file could not be parsed. Errors from '" + script + "' follow :\n" + open(errFile).read()


class CheckForBugs(plugins.Action):
    def __init__(self):
        self.diag = logging.getLogger("Check For Bugs")

    def callDuringAbandon(self, test):
        # want to be able to mark UNRUNNABLE tests as known bugs too...
        return test.state.lifecycleChange != "complete"

    def __repr__(self):
        return "Checking known bugs for"

    def __call__(self, test):
        newState, rerunCount = self.checkTest(test, test.state)
        if newState:
            test.changeState(newState)
        if rerunCount and not test.app.isReconnecting() and not os.path.exists(test.makeBackupFileName(rerunCount)):
            self.describe(test, " - found an issue that triggered a rerun")
            test.saveState()
            # Current thread, must be done immediately or we might exit...
            test.performNotify("Rerun")
            # for test synchronisation, mainly
            test.notify("RerunTriggered")

        if test.state.category == "killed" and os.path.exists(test.makeBackupFileName(1)):
            newState = test.restoreLatestBackup()
            if newState:
                self.fixBackupMessage(newState)
                test.changeState(newState)

    def checkTest(self, test, state):
        activeBugs = self.readBugs(test)
        return self.checkTestWithBugs(test, state, activeBugs)

    def checkTestWithBugs(self, test, state, activeBugs):
        if not activeBugs.checkUnchanged() and not state.hasFailed():
            self.diag.info(repr(test) + " succeeded, not looking for bugs")
            return None, 0

        bugTrigger, bugStem = self.findBug(test, state, activeBugs)
        if bugTrigger:
            if bugTrigger.bugInfo.rerunOnly:
                return None, bugTrigger.bugInfo.rerunCount
            else:
                absenceBug = bugTrigger in activeBugs[bugStem].absentList
                category, briefText, fullText = bugTrigger.findBugInfo(test, bugStem, absenceBug)
                self.diag.info("Changing to " + category + " with text " + briefText)
                bugState = FailedPrediction(category, fullText, briefText, completed=1)
                return self.getNewState(state, bugState), bugTrigger.bugInfo.rerunCount
        else:
            return None, 0

    def findAllBugs(self, test, state, activeBugs):
        multipleDiffs = self.hasMultipleDifferences(test, state)
        bugs, bugStems = [], []
        for stem, fileBugData in list(activeBugs.items()):
            newBugs = self.findBugsInFile(test, state, stem, fileBugData, multipleDiffs)
            if newBugs:
                bugs += newBugs
                bugStems += [stem] * len(newBugs)
        return bugs, bugStems

    def findBug(self, test, state, activeBugs):
        bugs, bugStems = self.findAllBugs(test, state, activeBugs)
        unblockedBugs = self.findUnblockedBugs(bugs)
        if len(unblockedBugs) > 0:
            unblockedBugs.sort(key=lambda bug: (bug.bugInfo.priority, bug.bugInfo.rerunCount))
            bug = unblockedBugs[0]
            return bug, bugStems[bugs.index(bug)]
        else:
            return None, None

    def findUnblockedBugs(self, bugs):
        unblockedBugs = []
        for bug in bugs:
            if bug.bugInfo.isCancellation():
                return unblockedBugs
            else:
                unblockedBugs.append(bug)
        return unblockedBugs

    def findBugsInFile(self, test, state, stem, fileBugData, multipleDiffs):
        self.diag.info("Looking for bugs in file " + stem)
        if stem == "free_text":
            return fileBugData.findBugsInText(state.freeText.split("\n"), execHosts=state.executionHosts, tmpDir=test.writeDirectory)
        elif stem == "brief_text":
            briefText = state.getTypeBreakdown()[1]
            return fileBugData.findBugsInText(briefText.split("\n"), execHosts=state.executionHosts, tmpDir=test.writeDirectory)
        elif state.hasResults():
            # bugs are only relevant if the file itself is changed, unless marked to trigger on success also
            bugs = []
            for comp in state.findComparisonsMatching(stem):
                isChanged = not comp.hasSucceeded()
                bugs += fileBugData.findBugs(comp.tmpFile, state.executionHosts, isChanged, multipleDiffs)
            return bugs
        return []

    def getNewState(self, oldState, bugState):
        if hasattr(oldState, "failedPrediction"):
            # if we've already compared, slot our things into the comparison object
            newState = copy(oldState)
            newState.setFailedPrediction(bugState, usePreviousText=True)
            return newState
        else:
            return bugState

    def hasMultipleDifferences(self, test, state):
        if not state.hasResults():
            # check for unrunnables...
            return False
        comparisons = state.getComparisons()
        diffCount = len(comparisons)
        if diffCount <= 1:
            return False
        perfStems = state.getPerformanceStems(test)
        for comp in comparisons:
            if comp.stem in perfStems:
                diffCount -= 1
        return diffCount > 1

    def readBugs(self, test):
        bugMap = BugMap()
        # Mostly for backwards compatibility, reverse the list so that more specific bugs
        # get checked first.
        for bugFile in reversed(test.getAllPathNames("knownbugs")):
            self.diag.info("Reading bugs from file " + bugFile)
            bugMap.readFromFile(bugFile)
        return bugMap

    def fixBackupMessage(self, newState):
        newFreeText = ""
        backupRegex = re.compile(Bug.rerunLine.replace("%d", "[0-9]*").replace("(", "\\("))
        for line in newState.freeText.splitlines():
            if backupRegex.match(line):
                newFreeText += "(NOTE: This issue triggered a rerun, but this rerun was killed before it could complete.\n" + \
                    "The results presented here are those of the completed run.\n"
            elif line == Bug.prevResultLine:
                newFreeText += line.replace("previous runs", "the killed rerun").replace("previous.*", "aborted") + "\n"
            else:
                newFreeText += line + "\n"
        newState.freeText = newFreeText


# For migrating from knownbugs files which are from TextTest 3.7 and older
class MigrateFiles(plugins.Action):
    def setUpSuite(self, suite):
        self.migrate(suite)

    def __call__(self, test):
        self.migrate(test)

    def __repr__(self):
        return "Migrating knownbugs file in"

    def migrate(self, test):
        for bugFileName in test.findAllStdFiles("knownbugs"):
            parser = ConfigParser()
            # Default behaviour transforms to lower case: we want case-sensitive
            parser.optionxform = str
            try:
                parser.read(bugFileName)
            except Exception:
                plugins.printWarning("Bug file at " + bugFileName + " not understood, ignoring")
                continue
            if not parser.has_section("Migrated section 1"):
                self.describe(test, " - " + os.path.basename(bugFileName))
                sys.stdout.flush()
                self.updateFile(bugFileName, parser)
            else:
                self.describe(test, " (already migrated)")

    def updateFile(self, bugFileName, parser):
        newBugFileName = bugFileName + ".new"
        newBugFile = open(newBugFileName, "w")
        self.writeNew(parser, newBugFile)
        newBugFile.close()
        print("Old File:\n" + open(bugFileName).read())
        print("New File:\n" + open(newBugFileName).read())
        shutil.move(newBugFileName, bugFileName)

    def writeNew(self, parser, newBugFile):
        sectionNo = 0
        for fileStem in parser.sections():
            for bugText in parser.options(fileStem):
                bugId = parser.get(fileStem, bugText)
                sectionNo += 1
                self.writeSection(newBugFile, sectionNo, fileStem, bugText, bugId)

    def writeSection(self, newBugFile, sectionNo, fileStem, bugText, bugId):
        newBugFile.write("[Migrated section " + str(sectionNo) + "]\n")
        newBugFile.write("search_string:" + bugText + "\n")
        newBugFile.write("search_file:" + fileStem + "\n")
        bugSystem = self.findBugSystem(bugId)
        if bugSystem:
            newBugFile.write("bug_system:" + bugSystem + "\n")
            newBugFile.write("bug_id:" + bugId + "\n")
        else:
            newBugFile.write("full_description:" + bugId + "\n")
            newBugFile.write("brief_description:unreported bug\n")
            newBugFile.write("internal_error:0\n")
        newBugFile.write("\n")

    def findBugSystem(self, bugId):
        for letter in bugId:
            if not letter in string.digits:
                return None
        return "bugzilla"




import os
import sys


def findLoggerNames(fileName, keyText="Logger"):
    result = []
    for line in open(fileName):
        if keyText in line:
            words = line.split('"')
            for i, word in enumerate(words):
                if word not in result and i % 2 == 1:  # Only take odd ones, between the quotes!
                    result.append(word)
    result.sort()
    return result


def findLoggerNamesUnder(location, **kwargs):
    result = set()
    for root, _, files in os.walk(location):
        for file in files:
            # Don't allow generation from ourselves...
            if file.endswith(".py") and file != "logconfiggen.py" and file != os.path.basename(sys.argv[0]):
                fileName = os.path.join(root, file)
                result.update(findLoggerNames(fileName, **kwargs))
    return sorted(result)


class PythonLoggingGenerator:
    def __init__(self, fileName, postfix="", prefix=""):
        self.file = open(fileName, "w")
        self.postfix = "." + postfix
        self.prefix = prefix
        self.handlers = {"stdout": "stdout"}

    def write(self, line):
        self.file.write(line + "\n")

    def parseInput(self, enabledLoggerNames, allLoggerNames):
        enabled, all = [], []
        for loggerInfo in enabledLoggerNames:
            try:
                loggerName, fileStem = loggerInfo
            except ValueError:
                loggerName = loggerInfo
                fileStem = loggerInfo
            enabled.append((loggerName, fileStem))
            all.append(loggerName)
        disabled = [l for l in allLoggerNames if l not in all]
        all += disabled
        return enabled, disabled, all

    def generate(self, enabledLoggerNames=[], allLoggerNames=[], debugLevelLoggers=[],
                 timeStdout=False, useDebug=True, defaultLevel="INFO"):
        enabled, disabled, all = self.parseInput(enabledLoggerNames, allLoggerNames)
        self.writeHeaderSections(timed=timeStdout)
        if len(enabled):
            self.write("# ====== The following are enabled by default ======")
            for loggerName, fileStem in enabled:
                self.writeLoggerSection(loggerName, True, fileStem, useDebug, defaultLevel)

        if len(disabled):
            self.write("# ====== The following are disabled by default ======")
            for loggerName in disabled:
                if loggerName in debugLevelLoggers:
                    level = "DEBUG"
                else:
                    level = defaultLevel
                self.writeLoggerSection(loggerName, False, loggerName, useDebug, level)
        self.writeFooterSections(all)

    def writeLoggerSection(self, loggerName, enable, fileStem, useDebug, level):
        self.write("# ======= Section for " + loggerName + " ======")
        self.write("[logger_" + loggerName + "]")
        handler = self.handlers.get(fileStem, loggerName)
        self.write("handlers=" + handler)
        self.write("qualname=" + loggerName)
        if enable:
            self.write("level=" + level + "\n")
        else:
            self.write("#level=" + level + "\n")

        if handler == loggerName:
            self.handlers[fileStem] = handler
            self.write("[handler_" + handler + "]")
            self.write("class=FileHandler")
            if enable or not useDebug:
                self.write("#formatter=timed")
            else:
                self.write("formatter=debug")
            fileName = self.prefix + fileStem.lower().replace(" ", "") + self.postfix
            if enable:
                self.write("#args=(os.devnull, 'a')")
                self.write("args=('" + fileName + "', 'a')\n")
            else:
                self.write("args=(os.devnull, 'a')")
                self.write("#args=('" + fileName + "', 'a')\n")

    def writeHeaderSections(self, timed=False):
        if timed:
            commentStr = ""
        else:
            commentStr = "#"
        self.write("""
[logger_root]
handlers=root
level=ERROR

[handler_root]
class=StreamHandler
level=ERROR
args=(sys.stdout,)

[handler_stdout]
class=StreamHandler
args=(sys.stdout,)
%sformatter=timed

[formatter_timed]
format=%%(asctime)s - %%(message)s

[formatter_debug]
format=%%(name)s %%(levelname)s - %%(message)s
""" % (commentStr))

    def writeFooterSections(self, loggerNames):
        loggerStr = ",".join(loggerNames)
        handlerStr = ",".join(sorted(set(self.handlers.values())))
        self.write("""# ====== Cruft that python logging module needs ======
[loggers]
keys=root,%s

[handlers]
keys=root,%s

[formatters]
keys=timed,debug
""" % (loggerStr, handlerStr))




import os
import string
import subprocess
from . import gridqueuesystem
from texttestlib.plugins import gethostname, log, TextTestError
from time import sleep
from locale import getpreferredencoding

# Used by master process for submitting, deleting and monitoring slave jobs


class QueueSystem(gridqueuesystem.QueueSystem):
    allStatuses = {"qw": ("PEND", "Pending"),
                   "hqw": ("HOLD", "On hold"),
                   "t": ("TRANS", "Transferring"),
                   "r": ("RUN", "Running"),
                   "s": ("USUSP", "Suspended by the user"),
                   "dRr": ("DEL", "In the process of being killed"),
                   "dr": ("DEL", "In the process of being killed"),
                   "dt": ("DEL", "In the process of being killed"),
                   "ds": ("DEL", "In the process of being killed"),
                   "dS": ("DEL", "In the process of being killed"),
                   "R": ("RESTART", "Restarted"),
                   "Rr": ("RESTART", "Restarted"),
                   "Rt": ("RESTART", "Restarted"),
                   "Rq": ("REQUEUED", "Requested a restart"),
                   "S": ("SSUSP", "Suspended by SGE due to other higher priority jobs"),
                   "St": ("SSUSP", "Suspended by SGE due to other higher priority jobs"),
                   "SR": ("SSUSP", "Suspended by SGE due to other higher priority jobs"),
                   "SRt": ("SSUSP", "Suspended by SGE due to other higher priority jobs"),
                   "T": ("THRESH", "Suspended by SGE as it exceeded allowed thresholds")}
    errorStatuses = ["Eqw", "ERq"]
    submitProg = "qsub"
    def __init__(self, *args):
        self.qdelOutput = ""
        self.errorReasons = {}
        gridqueuesystem.QueueSystem.__init__(self, *args)

    def getSlaveStartErrorFile(self):
        return os.path.join(self.coreFileLocation, "slave_start_errors." + os.getenv("USER"))

    def getSubmitCmdArgs(self, submissionRules, commandArgs=[], slaveEnv={}):
        qsubArgs = ["qsub", "-N", submissionRules.getJobName()]
        if submissionRules.processesNeeded != 1:
            qsubArgs += ["-pe", submissionRules.getParallelEnvironment(),
                         str(submissionRules.processesNeeded)]
        if submissionRules.useCoreBinding():
            qsubArgs += ["-binding", "linear:"+str(submissionRules.processesNeeded)]
        queue = submissionRules.findQueue()
        if queue:
            qsubArgs += ["-q", queue]
        priority = submissionRules.findPriority()
        if priority:
            qsubArgs += ["-p", str(priority)]
        resource = self.getResourceArg(submissionRules)
        if len(resource):
            qsubArgs += ["-l", resource]
        qsubArgs += ["-w", "e", "-notify", "-m", "n", "-cwd", "-b", "y"]
        if slaveEnv:
            if len(slaveEnv) >= len(os.environ):  # We've clearly copied the environment, just forward the whole thing
                qsubArgs.append("-V")
            else:
                qsubArgs.append("-v")
                qsubArgs.append(",".join(slaveEnv))

        qsubArgs += ["-o", os.devnull, "-e", self.getSlaveStartErrorFile()]
        return self.addExtraAndCommand(qsubArgs, submissionRules, commandArgs)

    def getResourceArg(self, submissionRules):
        resourceList = submissionRules.findResourceList()
        machines = submissionRules.findMachineList()
        if len(machines):
            resourceList.append("hostname=" + "|".join(machines))
        return ",".join(resourceList)

    def findSubmitError(self, stderr):
        return stderr.strip().splitlines()[0]

    def killJob(self, jobId):
        proc = subprocess.Popen(["qdel", jobId], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, encoding=getpreferredencoding())
        self.qdelOutput = proc.communicate()[0]
        return self.qdelOutput.find("has registered the job") != -1 or self.qdelOutput.find("has deleted job") != -1

    def setSuspendState(self, jobId, newState):
        arg = "-sj" if newState else "-usj"
        cmdArgs = ["qmod", arg, jobId]
        proc = subprocess.Popen(cmdArgs, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, encoding=getpreferredencoding())
        output = proc.communicate()[0]
        # unsuspend always provides return code 1, even when it works (bug in SGE)
        if newState and proc.returncode > 0:
            raise TextTestError("Failed to suspend job using command '" + " ".join(cmdArgs) +
                                "'\nError message from SGE follows:\n" + output)

    def getJobId(self, line):
        return line.split()[2]

    def findJobId(self, stdout):
        jobId = ""
        for line in stdout.splitlines():
            if line.find("has been submitted") != -1:
                jobId = self.getJobId(line)
            else:
                log.info("Unexpected output from qsub : " + line.strip())
        return jobId

    def getStatusForAllJobs(self):
        statusDict = {}
        proc = subprocess.Popen(["qstat"], stdin=open(os.devnull), stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding=getpreferredencoding())
        outMsg = proc.communicate()[0]
        if proc.returncode > 0:
            # SGE unavailable for the moment, don't update the job status
            return

        for line in outMsg.splitlines():
            words = line.split()
            if len(words) >= 5 and words[0].isdigit():
                jobId = words[0]
                statusLetter = self.getStatusLetter(words, 4)
                if statusLetter in self.errorStatuses:
                    self.errorReasons[jobId] = self.getErrorReason(jobId)
                    self.killJob(jobId)
                    continue

                status = self.allStatuses.get(statusLetter)
                if status:
                    statusDict[jobId] = status
                else:
                    log.info("WARNING: unexpected job status " + repr(statusLetter) + " received from SGE!")
                    statusDict[jobId] = statusLetter, statusLetter
        return statusDict

    def isDate(self, text):
        return len(text) == 10 and text.count("/") == 2

    def getStatusLetter(self, words, statusIndex):
        if len(words) < statusIndex + 1 or self.isDate(words[statusIndex + 1]):
            return words[statusIndex]
        else:
            return self.getStatusLetter(words, statusIndex + 1)

    def getErrorReason(self, jobId):
        proc = subprocess.Popen(["qstat", "-j", jobId], stdin=open(os.devnull), encoding=getpreferredencoding(),
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        outMsg = proc.communicate()[0]
        for line in outMsg.splitlines():
            if line.startswith("error reason"):
                return line.strip()
        return ""

    def getSlaveStartErrorMessage(self):
        errFile = self.getSlaveStartErrorFile()
        if not os.path.isfile(errFile):
            return ""

        errLines = open(errFile).read().splitlines()
        errText = "\n".join(errLines[-3:])
        if not errText:
            return ""

        return self.makeHeader("Recent errors written when starting SGE slave jobs") + errText + "\n(full file is at " + errFile + " - please remove this file sometime you aren't running TextTest)"

    def getJobFailureInfo(self, jobId):
        text = gridqueuesystem.QueueSystem.getJobFailureInfo(self, jobId)
        errors = self.getSlaveStartErrorMessage()
        if errors:
            text = errors + "\n" + text
        return text

    def _getJobFailureInfo(self, jobId):
        if jobId in self.errorReasons:
            return "SGE job entered error state: " + jobId + "\nTextTest terminated this job as a result. SGE's error reason follows:\n" + self.errorReasons.get(jobId)
        methods = [self.getAccountInfo, self.getAccountInfoOldFiles, self.retryAccountInfo]
        acctError = ""
        for method in methods:
            acctOutput, acctError = method(jobId)
            if acctOutput is not None:
                return acctOutput
        if self.qdelOutput:
            return "SGE lost job: " + jobId + "\nqdel output was as follows:\n" + self.qdelOutput
        else:
            return "Could not find info about job: " + jobId + "\nqacct error was as follows:\n" + acctError

    def getAccountInfo(self, jobId, extraArgs=[]):
        cmdArgs = ["qacct", "-j", jobId] + extraArgs
        proc = subprocess.Popen(cmdArgs, stdin=open(os.devnull), stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding=getpreferredencoding())
        outMsg, errMsg = proc.communicate()
        notFoundMsg = "error: job id " + jobId + " not found"
        if len(errMsg) == 0 or notFoundMsg not in errMsg:
            return outMsg, errMsg
        else:
            return None, errMsg

    def retryAccountInfo(self, jobId):
        sleepTime = 0.5
        acctError = ""
        for i in range(9):  # would be 10 but we had one already
            # assume failure is because the job hasn't propagated yet, wait a bit
            sleep(sleepTime)
            if sleepTime < 5:
                sleepTime *= 2
            acctOutput, acctError = self.getAccountInfo(jobId)
            if acctOutput is not None:
                return acctOutput, acctError
            else:
                log.info("Waiting " + str(sleepTime) + " seconds before retrying account info for job " + jobId)
        return None, acctError

    def getAccountInfoOldFiles(self, jobId):
        acctError = ""
        for logNum in range(5):
            # try at most 5 accounting files for now - assume jobs don't run longer than 5 days!
            fileName = self.findAccountingFile(logNum)
            if not fileName:
                return None, acctError
            acctInfo, acctError = self.getAccountInfo(jobId, ["-f", fileName])
            if acctInfo:
                return acctInfo, acctError
        return None, acctError

    def findAccountingFile(self, logNum):
        if "SGE_ROOT" in os.environ and "SGE_CELL" in os.environ:
            findPattern = os.path.join(os.environ["SGE_ROOT"], os.environ["SGE_CELL"])
            acctFile = os.path.join(findPattern, "common", "accounting." + str(logNum))
            if os.path.isfile(acctFile):
                return acctFile


# Used by slave for producing performance data
class MachineInfo:
    def findActualMachines(self, machineOrGroup):
        # In LSF this unpacks host groups, taking advantage of the fact that they are
        # interchangeable with machines. This is not true in SGE anyway, so don't support it.
        return [machineOrGroup]

    def findResourceMachines(self, resource):
        machines = []
        # Hacked workaround for problems with SGE, bug 1513 in their bug system
        # Should really use qhost but that seems flaky
        for line in os.popen("qselect -l '" + resource + "'"):
            fullMachine = line.strip().split("@")[-1]
            machineName = fullMachine.split(".")[0]
            if not machineName in machines:
                machines.append(machineName)
        return machines

    def findRunningJobs(self, machine):
        jobs = []
        user, jobId = "", ""
        myJobId = os.path.basename(os.getenv("SGE_JOB_SPOOL_DIR", "")).split(".")[0]
        for line in os.popen("qstat -r -s r -u '*' -l hostname='" + machine + "'"):
            if line.startswith("job") or line.startswith("----"):
                continue
            if line[0] in string.digits:
                fields = line.split()
                if fields[0] != myJobId:
                    user = fields[-6]
                    jobId = fields[0]
                else:
                    user, jobId = "", ""
            elif jobId and line.find("Full jobname") != -1:
                jobName = line.split(":")[-1].strip()
                jobs.append((user, jobId, jobName))
        return jobs

# Interpret what the limit signals mean...


def getUserSignalKillInfo(userSignalNumber, explicitKillMethod):
    if userSignalNumber == "1":
        return "RUNLIMIT", "exceeded maximum wallclock time allowed by SGE (s_rt parameter)"
    else:
        return explicitKillMethod()

# Used by slave to find all execution machines


def getExecutionMachines():
    hostFile = os.getenv("PE_HOSTFILE")
    if not hostFile or not os.path.isfile(hostFile):
        return [gethostname()]
    hostlines = open(hostFile).readlines()
    hostlist = []
    for line in hostlines:
        parts = line.strip().split()
        if len(parts) < 2:
            continue
        host = parts[0].split(".")[0]
        counter = int(parts[1])
        while counter > 0:
            hostlist.append(host)
            counter = counter - 1
    return hostlist




import os
import sys
import time
from texttestlib import plugins
from .comparefile import FileComparison

# This module won't work without an external module creating a file called performance.app
# This file should be of a format understood by the function below i.e. a single line containing
# CPU time   :      30.31 sec. on heathlands
#

# For memory the same is true, and the format is
# Max Memory :       45 MB

# Returns -1 as error value, if the file is the wrong format


def getPerformance(fileName):
    if not fileName:
        return float(-1)
    line = open(fileName).readline()
    return getPerformanceFromLine(line)


def getPerformanceFromLine(line):
    pos = line.find(":")
    if pos == -1:
        return float(-1)
    return float(line[pos + 1:].lstrip().split()[0])


def getTestPerformance(test, version=None):
    try:
        perfStem = test.getConfigValue("default_performance_stem")
        return getPerformance(test.getFileName(perfStem, version))
    except IOError:  # assume something disappeared externally
        test.refreshFiles()
        return getTestPerformance(test, version)


def describePerformance(fileName):
    line = open(fileName).readline().strip()
    if "mem" in os.path.basename(fileName):
        return line

    # Assume seconds
    perf = getPerformanceFromLine(line)
    description = getTimeDescription(perf)
    return line.replace(str(perf) + " sec.", description)


def getTimeDescription(seconds):
    values = list(time.gmtime(seconds)[2:6])
    values[0] -= 1  # not actually using timedelta which doesn't support formatting...
    units = ["day", "hour", "minute", "second"]
    parts = []
    for unit, val in zip(units, values):
        if val:
            parts.append(plugins.pluralise(val, unit))

    if len(parts) > 0:
        description = " and ".join(parts).replace(" and ", ", ", len(parts) - 2)
    else:
        description = "Less than 1 second"
    return description


def parseTimeExpression(timeExpression):
    # Starts with <, >, <=, >= ?
    if timeExpression == "":
        return "", 0
    if timeExpression.startswith("<="):
        return "<=", plugins.getNumberOfSeconds(timeExpression[2:])
    if timeExpression.startswith("<"):
        return "<", plugins.getNumberOfSeconds(timeExpression[1:])
    if timeExpression.startswith(">="):
        return ">=", plugins.getNumberOfSeconds(timeExpression[2:])
    if timeExpression.startswith(">"):
        return ">", plugins.getNumberOfSeconds(timeExpression[1:])
    else:
        raise plugins.TextTestError("Could not parse time expression '" + timeExpression +
                                    "' : all expressions must begin with '<' or '>'.")


class PerformanceConfigSettings:
    def __init__(self, test, stem):
        self.configName = self.getConfigName(stem)
        self.configMethod = test.getCompositeConfigValue

    def getConfigName(self, stem):
        if stem == "performance":
            return "cputime"
        else:
            return stem

    def flagSet(self, configEntry):
        return self.configMethod(configEntry, self.configName) == "true"

    def aboveMinimum(self, value, configEntry):
        minimum = self.configMethod(configEntry, self.configName)
        return value < 0 or value > minimum

    def getMinimumMultipleRange(self, value, configEntry):
        minimum = self.configMethod(configEntry, self.configName)
        if minimum == 0:
            return None, None
        multiple = int(value / minimum)
        return int(multiple * minimum), int((multiple + 1) * minimum)

    def getDescriptor(self, configEntry):
        desc = self._getDescriptor(configEntry, self.configName)
        if desc:
            return desc
        else:
            fallbackConfigName = self.getFallbackConfigName()
            postfix = "(" + self.configName + ")"
            return self._getDescriptor(configEntry, fallbackConfigName) + postfix

    def getFallbackConfigName(self):
        if "mem" in self.configName:
            return "memory"
        else:
            return "cputime"

    def ignoreImprovements(self):
        return self.configMethod("performance_ignore_improvements", self.configName) == "true"

    def _getDescriptor(self, configEntry, configName):
        fromConfig = self.configMethod(configEntry, configName)
        if len(fromConfig) > 0:
            name, briefDesc, longDesc = plugins.commasplit(fromConfig)
            plugins.addCategory(name, briefDesc, longDesc)
            return name


class PerformanceFileComparison(FileComparison):
    def __init__(self, *args, **kw):
        self.perfComparison = None
        FileComparison.__init__(self, *args, **kw)

    def cacheDifferences(self, test, testInProgress):
        # Don't allow process count of 0, which screws things up...
        if self.stdFile and self.tmpFile:
            oldPerf = getPerformance(self.stdFile)
            # If we didn't understand the old performance, overwrite it and behave like it didn't exist
            if (oldPerf == float(-1)):
                os.remove(self.stdFile)
                test.refreshFiles()
                self.stdFile = test.getFileName(self.stem)
                self.stdCmpFile = self.stdFile
                if not self.stdFile:
                    return

            newPerf = getPerformance(self.tmpFile)
            self.diag.info("Performance is " + str(oldPerf) + " and " + str(newPerf))
            settings = PerformanceConfigSettings(test, self.stem)
            self.perfComparison = PerformanceComparison(oldPerf, newPerf, settings)
            self.differenceCache = self.perfComparison.isSignificant(settings)

    def __repr__(self):
        baseText = FileComparison.__repr__(self)
        if self.newResult():
            return baseText
        return baseText + "(" + self.getType() + ")"

    def getType(self):
        if self.newResult():
            return FileComparison.getType(self)
        else:
            return self.perfComparison.descriptor

    def getDifferencesSummary(self, includeNumbers=True):
        return self.perfComparison.getSummary(includeNumbers)

    def getToleranceMultipleRange(self, test):
        settings = PerformanceConfigSettings(test, self.stem)
        return self.perfComparison.getToleranceMultipleRange(settings)

    def getDetails(self):
        if self.hasDifferences():
            return self.getDifferencesSummary()
        else:
            return ""

    def getNewPerformance(self):
        if self.perfComparison:
            return self.perfComparison.newPerformance
        else:
            return getPerformanceFromLine(self.freeTextBody.splitlines()[0])

    def saveResults(self, tmpFile, destFile):
        # Here we save the average of the old and new performance, assuming fluctuation
        avgPerformance = self.perfComparison.getAverage()
        self.diag.info("Found average performance = " + str(avgPerformance) +
                       ", new performance = " + str(self.perfComparison.newPerformance))
        line = open(tmpFile).readlines()[0]
        lineToWrite = line.replace(str(self.perfComparison.newPerformance), str(avgPerformance))
        newFile = open(destFile, "w")
        newFile.write(lineToWrite)

# class purely for comparing two performance numbers, independent of the files they come from


class PerformanceComparison:
    def __init__(self, oldPerf, newPerf, settings):
        self.oldPerformance = oldPerf
        self.newPerformance = newPerf
        self.percentageChange = self.calculatePercentageChange(settings)
        self.descriptor = self.getDescriptor(settings)

    def calculatePercentageChange(self, settings):
        if settings.flagSet("use_normalised_percentage_change"):
            return plugins.calculatePercentageNormalised(self.oldPerformance, self.newPerformance)
        else:
            return plugins.calculatePercentageStandard(self.oldPerformance, self.newPerformance)

    def getDescriptor(self, settings):
        if self.newPerformance < self.oldPerformance:
            return settings.getDescriptor("performance_descriptor_decrease")
        else:
            return settings.getDescriptor("performance_descriptor_increase")

    def getSummary(self, includeNumbers=True):
        if self.newPerformance < 0 or self.oldPerformance < 0:
            return "Performance comparison failed"

        perc = plugins.roundPercentage(self.percentageChange)
        if perc == 0:
            return ""
        elif perc == -1:
            return "infinitely " + self.descriptor
        elif includeNumbers:
            return str(perc) + "% " + self.descriptor
        else:
            return self.descriptor

    def getToleranceMultipleRange(self, settings):
        perc = plugins.roundPercentage(self.percentageChange)
        if perc <= 0:
            return ""
        lower, upper = settings.getMinimumMultipleRange(perc, "performance_variation_%")
        if lower is not None:
            return str(lower) + "%-" + str(upper) + "%"
        else:
            return ""

    def isSignificant(self, settings):
        if settings.ignoreImprovements() and self.newPerformance < self.oldPerformance:
            return False

        longEnough = settings.aboveMinimum(self.newPerformance, "performance_test_minimum") or \
            settings.aboveMinimum(self.oldPerformance, "performance_test_minimum")
        varianceEnough = settings.aboveMinimum(self.percentageChange, "performance_variation_%")
        return longEnough and varianceEnough

    def getAverage(self):
        return round((self.oldPerformance + self.newPerformance) / 2.0, 2)


class TimeFilter(plugins.Filter):
    option = "r"

    def __init__(self, timeLimit, *args):
        self.minTime = 0.0
        self.maxTime = sys.maxsize
        times = plugins.commasplit(timeLimit)
        if timeLimit.count("<") == 0 and timeLimit.count(">") == 0:  # Backwards compatible
            if len(times) == 1:
                self.maxTime = plugins.getNumberOfSeconds(timeLimit)
            else:
                self.minTime = plugins.getNumberOfSeconds(times[0])
                if len(times[1]):
                    self.maxTime = plugins.getNumberOfSeconds(times[1])
        else:
            for expression in times:
                parsedExpression = parseTimeExpression(expression)
                if parsedExpression[0] == "":
                    continue
                elif parsedExpression[0] == "<":
                    self.adjustMaxTime(parsedExpression[1] - 1)  # We don't care about fractions of seconds ...
                elif parsedExpression[0] == "<=":
                    self.adjustMaxTime(parsedExpression[1])
                elif parsedExpression[0] == ">":
                    self.adjustMinTime(parsedExpression[1] + 1)  # We don't care about fractions of seconds ...
                else:
                    self.adjustMinTime(parsedExpression[1])

    def adjustMinTime(self, newMinTime):
        if newMinTime > self.minTime:
            self.minTime = newMinTime

    def adjustMaxTime(self, newMaxTime):
        if newMaxTime < self.maxTime:
            self.maxTime = newMaxTime

    def acceptsTestCase(self, test):
        testPerformance = getTestPerformance(test)
        if testPerformance < 0:
            return True
        return testPerformance >= self.minTime and testPerformance <= self.maxTime


class TimeGroupFilter(plugins.Filter):
    def __init__(self, testCount, *args):
        self.testCount = testCount

    def makePerformanceDictionary(self, tests):
        dict = {}
        for test in tests:
            testPerformance = getTestPerformance(test)
            if testPerformance >= 0:
                dict.setdefault(testPerformance, set()).add(test)
        return dict

    def refine(self, tests):
        if self.testCount <= 0 or self.testCount >= len(tests):
            return tests

        testPerfDict = self.makePerformanceDictionary(tests)
        perfs = list(testPerfDict.keys())
        perfs.sort(self.comparePerformance)
        newTests = []
        for perf in perfs:
            for test in testPerfDict.get(perf):
                newTests.append(test)
                if len(newTests) == self.testCount:
                    return newTests
        return newTests


class FastestFilter(TimeGroupFilter):
    option = "fastest"

    def comparePerformance(self, perf1, perf2):
        return cmp(perf1, perf2)


class SlowestFilter(TimeGroupFilter):
    option = "slowest"

    def comparePerformance(self, perf1, perf2):
        return cmp(perf2, perf1)


class PerformanceStatistics(plugins.ScriptWithArgs):
    scriptDoc = "Prints a report on system resource usage per test. Can compare versions"
    printedTitle = False

    def __init__(self, args=[]):
        optDict = self.parseArguments(args, ["compv", "file"])
        self.settings = None
        self.compareVersion = optDict.get("compv")
        self.compareTotal = 0.0
        self.total = 0.0
        self.testCount = 0
        self.app = None
        self.file = optDict.get("file", "performance")

    def setUpSuite(self, suite):
        if suite.parent:
            print(suite.getIndent() + suite.name)
        else:
            entries = [suite.app.description(), "Version '" + self.app.getFullVersion() + "'"]
            if self.compareVersion is not None:
                entries += ["Version '" + self.compareVersion + "'", self.file + " change"]
            self.printUnderlined(self.getPaddedLine(entries))

    def printUnderlined(self, title):
        print("-" * len(title))
        print(title)
        print("-" * len(title))

    def getPaddedLine(self, entries):
        line = entries[0].ljust(40)
        for entry in entries[1:]:
            line += entry.rjust(20)
        return line

    def __call__(self, test):
        self.testCount += 1
        perf = getPerformance(test.getFileName(self.file))
        if perf > 0:
            self.total += perf
        entries = [test.getIndent() + test.name, self.format(perf)]
        if self.compareVersion is not None:
            comparePerf = getPerformance(test.getFileName(self.file, self.compareVersion))
            self.compareTotal += comparePerf
            self.settings = PerformanceConfigSettings(test, self.file)
            perfComp = PerformanceComparison(comparePerf, perf, self.settings)
            entries += [self.format(comparePerf), perfComp.getSummary()]
        print(self.getPaddedLine(entries))

    def format(self, number):
        if number < 0:
            return "N/A"
        if "mem" in self.file:
            return self.formatMemory(number)
        else:
            from datetime import timedelta
            return str(timedelta(seconds=int(number)))

    def formatMemory(self, memUsed):
        return str(memUsed) + " MB"

    def __del__(self):
        if not self.printedTitle:
            entries = ["Application/Version", "Total"]
            if self.compareVersion is not None:
                entries += ["Total (" + self.compareVersion + ")", self.file + " change"]
            entries.append("No. of Tests")
            self.printUnderlined(self.getPaddedLine(entries))
            PerformanceStatistics.printedTitle = True
        # Note - we might need to include parallel in this calculation...
        entries = [self.app.description(), self.format(self.total)]
        if self.compareVersion is not None:
            perfComp = PerformanceComparison(self.compareTotal, self.total, self.settings)
            entries += [self.format(self.compareTotal), perfComp.getSummary()]
        entries.append(str(self.testCount))
        print(self.getPaddedLine(entries))

    def setUpApplication(self, app):
        self.app = app




from . import abstractqueuesystem
import os, shutil
from texttestlib import plugins


class QueueSystem(abstractqueuesystem.QueueSystem):
    def __init__(self, app):
        if shutil.which(self.submitProg) is None:
            raise plugins.TextTestError("Cannot submit TextTest tests to grid engine: '" + self.submitProg + "' not installed!")
        self.coreFileLocation = self.getCoreFileLocation(app)

    def submitSlaveJob(self, cmdArgs, slaveEnv, logDir, *args, **kw):
        # Don't use log dir as working directory, it might not exist yet
        return abstractqueuesystem.QueueSystem.submitSlaveJob(self, cmdArgs, slaveEnv, self.coreFileLocation, *args, **kw)

    def getCoreFileLocation(self, app):
        location = app.getConfigValue("queue_system_core_file_location")
        if not location or not os.path.isdir(location):
            location = os.path.join(os.getenv("TEXTTEST_TMP"), "grid_core_files")
            plugins.ensureDirectoryExists(location)
        return location




# Generic interface to version control systems. We try to keep it as general as possible.
import os
import datetime
import subprocess
import shutil
import tempfile
from gi.repository import Gtk, Gdk, GObject
from texttestlib import plugins
from .. import guiplugins, guiutils, entrycompletion
from ..default_gui import adminactions, changeteststate
from . import custom_widgets

vcsClass, vcs, annotateClass, basicDiffClasses = None, None, None, None

# All VCS specific stuff goes in this class. One global instance, "vcs" above


class VersionControlInterface:
    def __init__(self, controlDir, name, warningStates, errorStates, latestRevisionName):
        self.name = name
        self.controlDirName = os.path.basename(controlDir)
        self.program = os.path.basename(controlDir).lower().replace(".", "")
        self.warningStates = warningStates
        self.errorStates = errorStates
        self.latestRevisionName = latestRevisionName
        self.lastMoveInVCS = False
        self.defaultArgs = {}

    def checkInstalled(self):
        self.callProgram("help")  # throws if it's not installed

    def isVersionControlled(self, path):
        basicArgs = self.getCmdArgs("status")
        for file in self.getFileNames(path, recursive=True, forStatus=True):
            status = self.getFileStatus(basicArgs, file)
            if status != "Unknown" and status != "Ignored":
                return True
        return False

    def getFileStatus(self, basicArgs, file):
        output = self.getProcessResults(basicArgs + [file])[1]
        return self.getStateFromStatus(output)

    def callProgram(self, cmdName, fileArgs=[], **kwargs):
        return subprocess.call(self.getCmdArgs(cmdName, fileArgs),
                               stdout=open(os.devnull, "w"), stderr=open(os.devnull, "w"), **kwargs)

    def callProgramOnFiles(self, cmdName, fileArg, recursive=False, extraArgs=[], **kwargs):
        basicArgs = self.getCmdArgs(cmdName, extraArgs)
        for fileName in self.getFileNamesForCmd(cmdName, fileArg, recursive):
            self.callProgramWithHandler(fileName, basicArgs + [fileName], **kwargs)

    def callProgramWithHandler(self, fileName, args, outputHandler=None, outputHandlerArgs=(), **kwargs):
        retcode, stdout, stderr = self.getProcessResults(args, **kwargs)
        if outputHandler:
            outputHandler(retcode, stdout, stderr, fileName, *outputHandlerArgs)

    def getProcessResults(self, args, **kwargs):
        try:
            process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                       universal_newlines=True, **kwargs)
        except OSError:
            raise plugins.TextTestError("Could not run " + self.name + ": make sure you have it installed locally")

        stdout, stderr = process.communicate()
        return process.returncode, stdout, stderr

    def getFileNamesForCmd(self, cmdName, fileArg, recursive):
        if cmdName == "add" and recursive:  # assume VCS adds recursively by default, override for CVS
            return [fileArg]
        else:
            return self.getFileNames(fileArg, recursive, forStatus=cmdName == "status")

    def getFileNames(self, fileArg, recursive, includeDirs=False, **kw):
        if os.path.isfile(fileArg):
            return [fileArg]
        elif os.path.isdir(fileArg):
            if includeDirs:
                baseFiles = [fileArg]
            else:
                baseFiles = []
            if recursive:
                return baseFiles + self.getFilesFromDirRecursive(fileArg, includeDirs)
            else:
                return baseFiles + self.getFilesFromDir(fileArg)
        else:
            return []

    def getFilesFromDir(self, dirName):
        files = []
        for f in sorted(os.listdir(dirName)):
            fullPath = os.path.join(dirName, f)
            if os.path.isfile(fullPath):
                files.append(fullPath)
        return files

    def getFilesFromDirRecursive(self, dirName, includeDirs):
        allFiles = []
        for root, dirs, files in os.walk(dirName):
            if self.controlDirName in dirs:
                dirs.remove(self.controlDirName)
            toAdd = files
            if includeDirs:
                toAdd += dirs
            for f in toAdd:
                fullPath = os.path.join(root, f)
                allFiles.append(fullPath)
            for dir in dirs:
                fullPath = os.path.join(root, dir)
                if os.path.islink(fullPath):
                    allFiles += self.getFilesFromDirRecursive(fullPath, includeDirs)

        return sorted(allFiles)

    def getProgramArgs(self):
        return [self.program]

    def getGraphicalDiffArgs(self, diffProgram):
        return [diffProgram]  # brittle but general...

    def getCmdArgs(self, cmdName, extraArgs=[]):
        return self.getProgramArgs() + [cmdName] + self.defaultArgs.get(cmdName, []) + extraArgs

    def getCombinedRevisionOptions(self, r1, r2):
        return ["-r", r1, "-r", r2]  # applies to CVS and Mercurial

    def copyPath(self, oldPath, newPath):
        if os.path.isdir(newPath):
            # After a remove, possibly, or after Mercurial has half-moved...
            for path in os.listdir(oldPath):
                oldSubPath = os.path.join(oldPath, path)
                newSubPath = os.path.join(newPath, path)
                if os.path.isdir(oldSubPath):
                    self.copyPath(oldSubPath, newSubPath)
                else:
                    shutil.copyfile(oldSubPath, newSubPath)
        else:
            plugins.copyPath(oldPath, newPath)

    def movePath(self, oldPath, newPath):
        self.lastMoveInVCS = self.isVersionControlled(oldPath)
        if self.lastMoveInVCS:
            if os.path.isdir(oldPath):
                newParent = os.path.dirname(newPath)
                # If it's also a parent of the old directory we don't need to check if it's version-controlled.
                if not oldPath.startswith(newParent) and not self.isVersionControlled(newParent):
                    self.callProgramOnFiles("add", newParent)
            self._movePath(oldPath, newPath)
        else:
            os.rename(oldPath, newPath)

    def _movePath(self, oldPath, newPath):
        self.callProgram("mv", [oldPath, newPath])

    def getMoveSuffix(self):
        if self.lastMoveInVCS:
            return " in " + self.name + " (using '" + self.getMoveCommand() + "')"
        else:
            return ""

    def getMoveCommand(self):
        return self.program + " mv"

    def hasLocalCommits(self, *args):
        return False  # Not possible in cvs or bzr

    def getRevisionOptions(self, r1, r2):
        if r1 and r2:
            return self.getCombinedRevisionOptions(r1, r2)
        elif r1:
            return self.getSingleRevisionOptions(r1)
        elif r2:
            return self.getSingleRevisionOptions(r2)
        else:
            return []

    def getSingleRevisionOptions(self, r1):
        return ["-r", r1]

    def startGUIProcess(self, cmdArgs, **kw):
        guiplugins.processMonitor.startProcess(cmdArgs, **kw)


class BasicVersionControlDialogGUI(guiplugins.ActionResultDialogGUI):
    recursive = False

    def getTitle(self, includeMnemonics=False, adjectiveAfter=True):
        title = self._getTitle()
        if self.recursive or not includeMnemonics:
            title = title.replace("_", "")

        if not includeMnemonics:
            # distinguish these from other actions that may have these names
            title = vcs.name + " " + title

        if self.recursive:
            title = self.add_recursion_annotation(title, adjectiveAfter)
        return title

    def add_recursion_annotation(self, title, adjectiveAfter):
        if adjectiveAfter:
            title += " Recursive"
        else:
            title = "Recursive " + title
        return title

    def createDialog(self):
        dialog = guiplugins.ActionResultDialogGUI.createDialog(self)
        dialog.set_name("VCS " + self._getTitle().replace("_", "") + " Window")
        return dialog

    def getDialogTitle(self):
        return self.getTitle(adjectiveAfter=False) + " for the selected files"

    def getTooltip(self):
        from copy import copy
        return copy(self.getDialogTitle()).replace(vcs.name, "version control").lower()


# Base class for all version control actions.
class VersionControlDialogGUI(BasicVersionControlDialogGUI):
    def __init__(self, allApps=[], dynamic=False, inputOptions={}):
        BasicVersionControlDialogGUI.__init__(self, allApps)
        self.cmdName = self._getTitle().replace("_", "").lower()
        self.dynamic = dynamic
        self.needsAttention = False
        self.notInRepository = False

    def showWarning(self):
        return self.notInRepository or self.needsAttention

    def getResultDialogIconType(self):
        if self.showWarning():
            return Gtk.STOCK_DIALOG_WARNING
        else:
            return Gtk.STOCK_DIALOG_INFO

    def getExtraArgs(self):
        return []

    def getFullResultTitle(self):
        return self.getResultTitle()

    def getResultDialogMessage(self):
        message = vcs.name + " " + self.getFullResultTitle() + " shown below."
        if self.needsAttention:
            message += "\n" + vcs.name + " " + self.getResultTitle() + " found files which are not up-to-date or which have conflicts"
        if self.notInRepository:
            message += "\nSome files/directories were not under " + vcs.name + " control."
        cmdArgs = vcs.getCmdArgs(self.cmdName, self.getExtraArgs())
        message += "\n" + vcs.name + " command used: " + " ".join(cmdArgs)
        if not self.recursive:
            message += "\nSubdirectories were ignored, use " + self.getTitle() + " Recursive to get the " + \
                self.getResultTitle() + " for all subdirectories."
        return message

    def extraResultDialogWidgets(self, exclude=""):
        all = ["log", "status", "diff", "annotate"]
        if self.cmdName in all:
            all.remove(self.cmdName)
        if exclude in all:
            all.remove(exclude)
        return all

    def commandHadError(self, retcode, *args):
        return retcode

    def outputIsInteresting(self, stdout):
        return True

    def getResultTitle(self):
        return self._getTitle().replace("_", "").lower()

    def runAndParse(self):
        self.notInRepository = False
        self.needsAttention = False
        extraArgs = self.getExtraArgs()
        for test, fileArg in self.getFilesForCmd():
            vcs.callProgramOnFiles(self.cmdName, fileArg, self.recursive, extraArgs,
                                   outputHandler=self.handleVcsOutput, outputHandlerArgs=(test,))

    def handleVcsOutput(self, retcode, stdout, stderr, fileName, test):
        if self.commandHadError(retcode, stderr, stdout):
            self.notInRepository = True
            self.storeResult(fileName, stderr, test)
        elif self.outputIsInteresting(stdout):
            self.storeResult(fileName, stdout, test)

    def storeResult(self, fileName, output, test):
        info = self.parseOutput(output)
        self.fileToTest[fileName] = test
        self.pages.append((fileName, output, info))
        dirName, local = os.path.split(fileName)
        self.notify("Status", "Analyzing " + self.getResultTitle() +
                    " for " + local + " in test " + os.path.basename(dirName))
        self.notify("ActionProgress", "")

    def parseOutput(self, output):
        return ""

    def updateSelection(self, *args):
        newActive = BasicVersionControlDialogGUI.updateSelection(self, *args)
        if not self.dynamic:  # See bugzilla 17653
            self.currFileSelection = []
        return newActive

    def isActiveOnCurrent(self, *args):
        return (len(self.currTestSelection) > 0 or len(self.currFileSelection) > 0) and not self.newFilesOnly()

    def newFilesOnly(self):
        return len(self.currFileSelection) > 0 and all((comp and hasattr(comp, "newResult") and comp.newResult() for f, comp in self.currFileSelection))

    def messageAfterPerform(self):
        return "Performed " + self.getTooltip() + "."

    def getResultDialogTwoColumnsInTreeView(self):
        return False

    def getResultDialogSecondColumnTitle(self):
        return "Information"

    def getSelectedFile(self):
        return self.filteredTreeModel.get_value(self.treeView.get_selection().get_selected()[1], 3)

    def viewStatus(self, button):
        file = self.getSelectedFile()
        self.diag.info("Viewing status on file " + file)
        status = StatusGUI()
        self.runWithSelections(status, file)

    def getFileApps(self, file):
        if len(self.currFileSelection) == 0:
            return []
        for f, apps in self.currFileSelection:
            if f == file:
                return apps
        return self.currFileSelection[0][1]

    def runWithSelections(self, gui, file):
        gui.notifyTopWindow(self.topWindow)
        test = self.fileToTest[file]
        gui.currTestSelection = [test] if test else []
        gui.currFileSelection = [(file, self.getFileApps(file))]
        gui.performOnCurrent()

    def viewLog(self, button):
        file = self.getSelectedFile()
        logger = LogGUI(self.validApps, self.dynamic)
        self.runWithSelections(logger, file)

    def viewAnnotations(self, button):
        file = self.getSelectedFile()
        annotater = annotateClass()
        self.runWithSelections(annotater, file)

    def viewDiffs(self, button, differ):
        file = self.getSelectedFile()
        differ.setRevisions(self.revisionEntry1.get_text(), self.revisionEntry2.get_text())
        self.runWithSelections(differ, file)

    def viewGraphicalDiff(self, button):
        path = self.filteredTreeModel.get_value(self.treeView.get_selection().get_selected()[1], 3)
        pathStem = os.path.basename(path).split(".")[0]
        diffProgram = guiutils.guiConfig.getCompositeValue("diff_program", pathStem)
        try:
            cmdArgs, exitHandlerArgs = self.getGraphicalDiffProgramArgs(diffProgram, path)
            if not cmdArgs:
                return  # already shown error
            vcs.startGUIProcess(cmdArgs, description="Graphical " + vcs.name + " diff for file " + path,
                                exitHandler=self.diffingComplete, exitHandlerArgs=exitHandlerArgs,
                                stderr=open(os.devnull, "w"), stdout=open(os.devnull, "w"))
        except OSError:
            self.showErrorDialog("\nCannot find graphical " + vcs.name + " difference program '" + diffProgram +
                                 "'.\nPlease install it somewhere on your $PATH.\n")

    def getGraphicalDiffProgramArgs(self, diffProgram, path):
        revOptions = self.getExtraArgs()
        graphDiffArgs = vcs.getGraphicalDiffArgs(diffProgram)
        if not graphDiffArgs[0] == diffProgram:
            # In order to get the correct error message if it doesn't exist
            subprocess.call([diffProgram, "--help"], stderr=open(os.devnull, "w"), stdout=open(os.devnull, "w"))
        return graphDiffArgs + revOptions + [path], ()

    def diffingComplete(self, *args):
        self.applicationEvent("the version-control graphical diff program to terminate")

    def getRootPath(self):
        if len(self.currTestSelection) > 0:
            app = self.currTestSelection[0].app
        else:
            app = self.currFileSelection[0][1][0]
        appPath = app.getDirectory()
        return os.path.split(appPath.rstrip(os.sep))[0]

    def getFilesForCmd(self):
        if len(self.currTestSelection) == 0:
            return [(None, f) for (f, _) in self.currFileSelection]
        elif len(self.currFileSelection) > 0:
            return [(self.currTestSelection[0], f) for (f, comp) in self.currFileSelection]
        else:
            testLists = [self.getFilesForCmdForTest(test) for test in self.currTestSelection]
            return sum(testLists, [])

    def getFilesForCmdForTest(self, test):
        testPath = test.getDirectory()
        if self.dynamic:
            files = sorted([fileComp.stdFile for fileComp in self.getComparisons(test)])
            return [(test, f) for f in files]
        else:
            return [(test, testPath)]

    def getComparisons(self, test):
        try:
            # Leave out new ones
            return test.state.changedResults + test.state.correctResults + test.state.missingResults
        except AttributeError:
            raise plugins.TextTestError("Cannot establish which files should be compared as no comparison information exists.\n" +
                                        "To create this information, perform 'recompute status' (press '" +
                                        guiutils.guiConfig.getCompositeValue("gui_accelerators", "recompute_status") + "') and try again.")

    def isModal(self):
        return False

    def addContents(self):
        self.pages = []
        self.fileToTest = {}
        self.runAndParse()  # will write to the above two structures
        self.vbox = Gtk.VBox()
        self.addExtraWidgets()
        self.addHeader()
        self.addTreeView()

    def addExtraWidgets(self):
        self.extraWidgetArea = Gtk.HBox()
        self.extraButtonArea = Gtk.HButtonBox()
        self.extraWidgetArea.pack_start(self.extraButtonArea, False, False, 0)
        if len(self.pages) > 0:
            padding = Gtk.Alignment.new(0.0, 0.0, 0.0, 0.0)
            padding.set_padding(3, 3, 3, 3)
            padding.add(self.extraWidgetArea)
            self.dialog.vbox.pack_end(padding, False, False, 0)
            extraWidgetsToShow = self.extraResultDialogWidgets()
            if "status" in extraWidgetsToShow:
                self.addStatusWidget()
            if "log" in extraWidgetsToShow:
                self.addLogWidget()
            if "annotate" in extraWidgetsToShow:
                self.addAnnotateWidget()
            if "graphical_diff" in extraWidgetsToShow:
                self.addGraphicalDiffWidget()
            if "diff" in extraWidgetsToShow:
                self.addDiffWidget()

    def addStatusWidget(self):
        button = Gtk.Button("_Status")
        button.connect("clicked", self.viewStatus)
        self.extraButtonArea.pack_start(button, False, False, 0)

    def addLogWidget(self):
        button = Gtk.Button("_Log")
        button.connect("clicked", self.viewLog)
        self.extraButtonArea.pack_start(button, False, False, 0)

    def addAnnotateWidget(self):
        button = Gtk.Button("_Annotate")
        button.connect("clicked", self.viewAnnotations)
        self.extraButtonArea.pack_start(button, False, False, 0)

    def addDiffWidget(self):
        label1 = Gtk.Label(label=" between revisions ")
        label2 = Gtk.Label(label=" and ")
        self.revisionEntry1 = Gtk.Entry()
        self.revisionEntry1.set_name("Revision 1")
        entrycompletion.manager.register(self.revisionEntry1)
        self.revisionEntry1.set_text(vcs.latestRevisionName)
        self.revisionEntry2 = Gtk.Entry()
        self.revisionEntry2.set_name("Revision 2")
        entrycompletion.manager.register(self.revisionEntry2)
        self.revisionEntry1.set_alignment(1.0)
        self.revisionEntry2.set_alignment(1.0)
        self.revisionEntry1.set_width_chars(6)
        self.revisionEntry2.set_width_chars(6)
        for diffClass in basicDiffClasses:
            diffObj = diffClass()
            diffButton = Gtk.Button(diffObj._getTitle() + "s")
            diffButton.connect("clicked", self.viewDiffs, diffObj)
            self.extraButtonArea.pack_start(diffButton, False, False, 0)
        self.extraWidgetArea.pack_start(label1, False, False, 0)
        self.extraWidgetArea.pack_start(self.revisionEntry1, False, False, 0)
        self.extraWidgetArea.pack_start(label2, False, False, 0)
        self.extraWidgetArea.pack_start(self.revisionEntry2, False, False, 0)

    def addGraphicalDiffWidget(self):
        button = Gtk.Button("_Graphical Diffs")
        button.connect("clicked", self.viewGraphicalDiff)
        self.extraButtonArea.pack_start(button, False, False, 0)

    def addHeader(self):
        message = self.getResultDialogMessage()
        if message:
            hbox = Gtk.HBox()
            iconType = self.getResultDialogIconType()
            hbox.pack_start(self.getStockIcon(iconType), False, False, 0)
            hbox.pack_start(Gtk.Label(message), False, False, 0)
            alignment = Gtk.Alignment.new(0.0, 1.0, 1.0, 1.0)
            alignment.set_padding(5, 5, 0, 5)
            alignment.add(hbox)
            self.vbox.pack_start(alignment, False, False, 0)

    def getStockIcon(self, stockItem):
        imageBox = Gtk.VBox()
        imageBox.pack_start(Gtk.Image.new_from_stock(stockItem, Gtk.IconSize.DIALOG), False, True, 0)
        return imageBox

    def addTreeView(self):
        hpaned = Gtk.HPaned()
        hpaned.set_name("VCS dialog separator")  # Mostly so we can filter the proportions, which we don't set

        # We need buffer when creating treeview, so create right-hand side first ...
        self.textBuffer = Gtk.TextBuffer()
        textView = Gtk.TextView.new_with_buffer(self.textBuffer)
        textView.set_editable(False)
        textView.set_name("VCS Output View")
        window2 = Gtk.ScrolledWindow()
        window2.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        window2.add(textView)
        hpaned.pack2(window2, True, True)

        self.createTreeView()
        window1 = Gtk.ScrolledWindow()
        window1.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        window1.add(self.treeView)
        hpaned.pack1(window1, False, True)

        if len(self.pages) > 0:
            parentSize = self.topWindow.get_size()
            self.dialog.resize(parentSize[0], int(parentSize[0] / 1.5))
            self.vbox.pack_start(hpaned, True, True, 0)
        self.dialog.vbox.pack_start(self.vbox, True, True, 0)

    def createTreeView(self):
        # Columns are: 0 - Tree node name
        #              1 - Content (output from VCS) for the corresponding file
        #              2 - Info. If the plugin wants to show two columns, this
        #                  is shown in the second column. If not it should be empty.
        #              3 - Full path to the file corresponding to the node
        #              4 - Should the row be visible?
        self.treeModel = Gtk.TreeStore(GObject.TYPE_STRING, GObject.TYPE_STRING,
                                       GObject.TYPE_STRING, GObject.TYPE_STRING,
                                       GObject.TYPE_BOOLEAN)
        self.filteredTreeModel = self.treeModel.filter_new()
        self.filteredTreeModel.set_visible_column(4)
        rootDir = self.getRootPath()
        fileToIter = {}
        for fileName, content, info in self.pages:
            label = plugins.relpath(fileName, rootDir)
            self.diag.info("Adding info for file " + label)
            path = label.split(os.sep)
            currentFile = rootDir
            prevIter = None
            for element in path:
                currentFile = os.path.join(currentFile, element)
                currentInfo = ""
                currentElement = element.strip(" \n")
                if currentFile == fileName:
                    currentInfo = info
                else:
                    currentElement = "<span weight='bold'>" + currentElement + "</span>"
                currIter = fileToIter.get(currentFile)
                if currIter is None:
                    newRow = (currentElement, content, currentInfo, currentFile, True)
                    currIter = self.treeModel.append(prevIter, newRow)
                    fileToIter[currentFile] = currIter
                prevIter = currIter

        self.treeView = Gtk.TreeView(self.filteredTreeModel)
        self.treeView.set_name("VCS " + self.cmdName + " info tree")
        self.treeView.set_enable_search(False)
        fileRenderer = Gtk.CellRendererText()
        fileColumn = Gtk.TreeViewColumn("File", fileRenderer, markup=0)
        fileColumn.set_resizable(True)
        self.treeView.append_column(fileColumn)
        self.treeView.set_expander_column(fileColumn)
        if self.getResultDialogTwoColumnsInTreeView():
            infoRenderer = Gtk.CellRendererText()
            self.infoColumn = custom_widgets.ButtonedTreeViewColumn(
                self.getResultDialogSecondColumnTitle(), infoRenderer, markup=2)
            self.infoColumn.set_resizable(True)
            self.treeView.append_column(self.infoColumn)
        self.treeView.get_selection().set_select_function(self.canSelect, self)
        self.treeView.expand_all()

        if len(self.pages) > 0:
            firstFile = self.pages[0][0]
            iterValid, firstIter = self.filteredTreeModel.convert_child_iter_to_iter(fileToIter[firstFile])
            self.updateForIter(firstIter)
            self.treeView.get_selection().select_iter(firstIter)

        self.treeView.get_selection().connect("changed", self.showOutput)

    def updateForIter(self, iter):
        self.extraWidgetArea.set_sensitive(True)
        text = self.filteredTreeModel.get_value(iter, 1)
        self.textBuffer.set_text(text)
        return text

    def showOutput(self, selection):
        model, iter = selection.get_selected()
        if iter:
            self.updateForIter(iter)
        else:
            self.extraWidgetArea.set_sensitive(False)

    @staticmethod
    def canSelect(selection, model, path, is_selected, user_data):
        return not user_data.treeModel.iter_has_child(
            user_data.treeModel.get_iter(user_data.filteredTreeModel.convert_path_to_child_path(path)))


#
# 1 - First the methods which just check the repository and checked out files.
#


class LogGUI(VersionControlDialogGUI):
    def _getTitle(self):
        return "_Log"

    def getResultTitle(self):
        return "logs"

    def getResultDialogTwoColumnsInTreeView(self):
        return True

    def getResultDialogSecondColumnTitle(self):
        return "Last revision committed (UTC)"

    def parseOutput(self, output):
        then = vcs.getDateFromLog(output)
        if then is None:
            return "Not in " + vcs.name

        now = datetime.datetime.utcnow()
        return self.getTimeDifference(now, then)

    # Show a human readable time difference string. Diffs larger than farAwayLimit are
    # written as the actual 'to' time, while other diffs are written e.g. 'X days ago'.
    # If markup is True, diffs less than closeLimit are boldified and diffs the same
    # day are red as well.
    def getTimeDifference(self, now, then, markup=True,
                          closeLimit=datetime.timedelta(days=3),
                          farAwayLimit=datetime.timedelta(days=7)):
        difference = now - then  # Assume this is positive ...
        if difference > farAwayLimit:
            return then.ctime()

        stringDiff = str(difference.days) + " days ago"
        yesterday = now - datetime.timedelta(days=1)
        if now.day == then.day:
            stringDiff = "Today at " + then.strftime("%H:%M:%S")
            if markup:
                stringDiff = "<span weight='bold' foreground='red'>" + stringDiff + "</span>"
        elif yesterday.day == then.day and yesterday.month == then.month and yesterday.year == then.year:
            stringDiff = "Yesterday at " + then.strftime("%H:%M:%S")
            if markup:
                stringDiff = "<span weight='bold'>" + stringDiff + "</span>"
        elif difference <= closeLimit and markup:
            stringDiff = "<span weight='bold'>" + stringDiff + "</span>"
        return stringDiff


class GenericDiffGUI(VersionControlDialogGUI):
    def __init__(self, *args):
        VersionControlDialogGUI.__init__(self, *args)
        self.revision1 = ""
        self.revision2 = ""

    def setRevisions(self, rev1, rev2):
        self.revision1 = rev1
        self.revision2 = rev2

    def getResultDialogMessage(self):
        if len(self.pages) == 0:
            return "All files are up-to-date and unmodified compared to the latest repository version."
        else:
            return VersionControlDialogGUI.getResultDialogMessage(self)

    def getFullResultTitle(self):
        return self.getResultTitle() + " " + self.getRevisionMessage()

    def showWarning(self):
        return len(self.pages) > 0

    def getRevisionMessage(self):
        if self.revision1 == "" and self.revision2 == "":
            return "compared to the latest revision"
        elif self.revision1 == "":
            return "between the local file and revision " + self.revision2
        elif self.revision2 == "":
            return "between revision " + self.revision1 + " and the local file"
        else:
            return "between revisions " + self.revision1 + " and " + self.revision2

    def extraResultDialogWidgets(self):
        return VersionControlDialogGUI.extraResultDialogWidgets(self, "diff") + ["graphical_diff"]


class DiffGUI(GenericDiffGUI):
    def __init__(self, *args):
        GenericDiffGUI.__init__(self, *args)
        self.cmdName = "diff"

    def _getTitle(self):
        return "_Difference"

    def getResultTitle(self):
        return "differences"

    def getExtraArgs(self):
        return vcs.getRevisionOptions(self.revision1, self.revision2)

    def commandHadError(self, retcode, stderr, *args):
        # Diff returns an error code for differences, not just for errors
        return retcode and len(stderr) > 0

    def outputIsInteresting(self, stdout):
        # Don't show diffs if they're empty
        return len(stdout) > 0


class FilteredDiffGUI(GenericDiffGUI):
    def __init__(self, *args):
        GenericDiffGUI.__init__(self, *args)
        self.cmdName = "cat"

    def _getTitle(self):
        return "_Filtered Difference"

    def getResultTitle(self):
        return "filtered differences"

    def getTmpFileArgs(self, fileName, revision):
        extraArgs = vcs.getSingleRevisionOptions(revision) if revision else []
        extraArgs.append(fileName)
        return extraArgs

    def makeTmpFile(self, fileName, revision, tmpDir):
        extraArgs = self.getTmpFileArgs(fileName, revision)
        args = vcs.getCmdArgs(self.cmdName, extraArgs)
        retcode, stdout, stderr = vcs.getProcessResults(args)
        if self.commandHadError(retcode, stderr, stdout):
            return None, stderr

        return self.writeTmpFile(stdout, tmpDir, os.path.basename(fileName)), None

    def writeTmpFile(self, text, tmpDir, fileName):
        path = os.path.join(tmpDir, fileName)
        with open(path, "w") as f:
            f.write(text)
        return path

    def runAndParse(self):
        self.notInRepository = False
        self.needsAttention = False
        for test, fileArg in self.getFilesForCmd():
            diffProgram = test.getConfigValue("text_diff_program")
            tmpDir = tempfile.mkdtemp()
            for fileName in vcs.getFileNamesForCmd("diff", fileArg, self.recursive):
                args, errors = self.getDiffProgramArgs(diffProgram, fileName, test, tmpDir)
                if errors:
                    self.storeResult(fileName, errors, test)
                    continue
                proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
                output = proc.communicate()[0]
                self.storeResult(fileName, output, test)
            shutil.rmtree(tmpDir)

    def getDiffProgramArgs(self, diffProgram, fileName, test, tmpDir):
        dir1 = os.path.join(tmpDir, "revision" + self.revision1 if self.revision1 else vcs.name)
        if not os.path.isdir(dir1):
            os.mkdir(dir1)
        file1, errors = self.makeTmpFile(fileName, self.revision1, dir1)
        if errors:
            return None, errors

        dir2 = os.path.join(tmpDir, "revision" + self.revision2 if self.revision2 else "local")
        if not os.path.isdir(dir2):
            os.mkdir(dir2)
        if self.revision2:
            file2, errors = self.makeTmpFile(fileName, self.revision2, dir2)
            if errors:
                return None, errors
        else:
            file2 = fileName
        filterName = os.path.basename(fileName) + "(FILTERED)"
        filteredFile1 = self.writeTmpFile(test.app.applyFiltering(test, file1), dir1, filterName)
        filteredFile2 = self.writeTmpFile(test.app.applyFiltering(test, file2), dir2, filterName)
        return [diffProgram, filteredFile1, filteredFile2], None

    def getGraphicalDiffProgramArgs(self, diffProgram, path):
        test = self.fileToTest[path]
        tmpDir = tempfile.mkdtemp()
        args, errors = self.getDiffProgramArgs(diffProgram, path, test, tmpDir)
        if errors:
            self.showErrorDialog("\nCannot show graphical differences. Could not find given files or revisions.\n")
        return args, (tmpDir,)

    def diffingComplete(self, tmpDir):
        GenericDiffGUI.diffingComplete(self, tmpDir)
        shutil.rmtree(tmpDir)


class StatusGUI(VersionControlDialogGUI):
    def __init__(self, *args):
        VersionControlDialogGUI.__init__(self, *args)
        self.popupMenu = None

    def _getTitle(self):
        return "_Status"

    def getResultDialogTwoColumnsInTreeView(self):
        return True

    def getStatusMarkup(self, status):
        if status in vcs.warningStates:
            return "<span weight='bold'>" + status + "</span>"
        elif status in vcs.errorStates:
            return "<span weight='bold' foreground='red'>" + status + "</span>"
        else:
            return status

    def parseOutput(self, output):
        status = vcs.getStateFromStatus(output)
        if status == "Unknown":
            self.notInRepository = True
        elif status in vcs.errorStates:
            self.needsAttention = True
        return self.getStatusMarkup(status)

    def createPopupMenu(self):
        # Each unique info column (column 2) gets its own menu item in the popup menu
        uniqueInfos = []
        self.treeModel.foreach(self.collectInfos, uniqueInfos)

        menu = Gtk.Menu()
        for info in uniqueInfos:
            menuItem = Gtk.CheckMenuItem(info)
            menuItem.set_active(True)
            menuItem.set_name(info)
            menu.append(menuItem)
            menuItem.connect("toggled", self.toggleVisibility)
            menuItem.show()
        return menu

    def toggleVisibility(self, menuItem):
        self.treeModel.foreach(self.setVisibility, (menuItem.get_name(), menuItem.get_active()))
        self.treeView.expand_row(self.filteredTreeModel.get_path(self.filteredTreeModel.get_iter_first()), True)

    def getStatus(self, iter):
        markedUpStatus = self.treeModel.get_value(iter, 2)
        start = markedUpStatus.find(">")
        if start == -1:
            return markedUpStatus
        else:
            end = markedUpStatus.rfind("<")
            return markedUpStatus[start + 1:end]

    def setVisibility(self, model, path, iter, xxx_todo_changeme):
        (actionName, actionState) = xxx_todo_changeme
        if model.iter_parent(iter) is not None and (actionName == "" or self.getStatus(iter) == actionName):
            model.set_value(iter, 4, actionState)
            parentIter = model.iter_parent(iter)
            if actionState or self.hasNoVisibleChildren(model, parentIter):
                self.setVisibility(model, model.get_path(parentIter), parentIter, ("", actionState))

    def hasNoVisibleChildren(self, model, iter):
        i = model.iter_children(iter)
        while i:
            if model.get_value(i, 4):
                return False
            i = model.iter_next(i)
        return True

    def collectInfos(self, model, path, iter, infos):
        info = model.get_value(iter, 2)
        if info != "":
            rawInfo = info.replace("<span weight='bold'>", "").replace("<span weight='bold' foreground='red'>",
                                                                       "").replace("</span>", "").strip()
            if rawInfo not in infos:
                infos.append(rawInfo)

    def addContents(self):
        VersionControlDialogGUI.addContents(self)
        self.popupMenu = self.createPopupMenu()
        self.infoColumn.set_clickable(True)
        self.infoColumn.get_button().connect("button-press-event", self.showPopupMenu)
        self.treeView.grab_focus()  # Or the column button gets focus ...

    def showPopupMenu(self, button, event):
        if event.button == 3:
            self.popupMenu.popup(None, None, None, event.button, event.time)


class AnnotateGUI(VersionControlDialogGUI):
    def _getTitle(self):
        return "A_nnotate"

    def getResultTitle(self):
        return "annotations"


class UpdateGUI(BasicVersionControlDialogGUI):
    @staticmethod
    def _getTitle():
        return "Update"

    def getSignalsSent(self):
        return ["Refresh"]

    def addContents(self):
        vcsDirectory = self.currTestSelection[0].app.getDirectory(
        ) if self.currTestSelection else self.validApps[0].getDirectory()
        canUpdate = not vcs.hasLocalCommits(vcsDirectory)
        if canUpdate:
            args = vcs.getCmdArgs(self.getCommandName())
            self.notify("Status", "Updating from " + vcs.name)
            self.notify("ActionStart")
            _, stdout, stderr = vcs.getProcessResults(args, cwd=vcsDirectory)
            text = stdout + stderr

        else:
            text = "You have local commits. Aborting updating via TextTest. You will need to merge by hand."
        buffer = Gtk.TextBuffer()
        buffer.set_text(text)
        textView = Gtk.TextView.new_with_buffer(buffer)
        self.dialog.vbox.pack_start(textView, True, True, 0)
        if canUpdate:
            self.notify("Refresh")
            self.notify("ActionStop")

    def getCommandName(self):
        return "update"

    def isActiveOnCurrent(self, *args):
        return self.shouldShow()


class AddGUI(VersionControlDialogGUI):
    def _getTitle(self):
        return "A_dd"

    def getResultDialogMessage(self):
        message = "Output from '" + vcs.name + " add' shown below."
        if not self.recursive:
            message += "\nSubdirectories were ignored, use " + self.getTitle() + " Recursive to add the files from all subdirectories."
        return message

    def commandHadError(self, retcode, stderr, *args):
        # Particularly CVS likes to write add output on stderr for some reason...
        return len(stderr) > 0


class VcsAdminAction:
    @staticmethod
    def removePath(*args):
        return vcs.removePath(*args)

    @staticmethod
    def movePath(*args):
        return vcs.movePath(*args)

    @staticmethod
    def copyPath(*args):
        return vcs.copyPath(*args)


class VcsRemoveAction(VcsAdminAction):
    def getFileRemoveWarning(self):
        return "Any " + vcs.name + "-controlled files will be removed in " + vcs.name + ".\n" + \
               "Any files that are not version controlled will be removed from the file system and hence may not be recoverable."


class RemoveTests(VcsRemoveAction, adminactions.RemoveTests):
    pass


class RemoveFiles(VcsRemoveAction, adminactions.RemoveFiles):
    pass


class RemoveTestsForPopup(VcsRemoveAction, adminactions.RemoveTestsForPopup):
    pass


class FindKnownBugs(VcsAdminAction, changeteststate.FindKnownBugs):
    pass


class RenameTest(VcsAdminAction, adminactions.RenameTest):
    def getNameChangeMessage(self, newName):
        origMessage = adminactions.RenameTest.getNameChangeMessage(self, newName)
        return origMessage + vcs.getMoveSuffix()


class RenameFile(VcsAdminAction, adminactions.RenameFile):
    def getNameChangeMessage(self, newName):
        origMessage = adminactions.RenameFile.getNameChangeMessage(self, newName)
        return origMessage + vcs.getMoveSuffix()


class PasteTests(VcsAdminAction, adminactions.PasteTests):
    def getStatusMessage(self, *args):
        origMessage = adminactions.PasteTests.getStatusMessage(self, *args)
        if self.removeAfter:
            return origMessage + vcs.getMoveSuffix()
        else:
            return origMessage


class LogGUIRecursive(LogGUI):
    recursive = True


class DiffGUIRecursive(DiffGUI):
    recursive = True


class FilteredDiffGUIRecursive(FilteredDiffGUI):
    recursive = True


class StatusGUIRecursive(StatusGUI):
    recursive = True


class AnnotateGUIRecursive(AnnotateGUI):
    recursive = True


class AddGUIRecursive(AddGUI):
    recursive = True

#
# Configuration for the Interactive Actions
#


class InteractiveActionConfig(guiplugins.InteractiveActionConfig):
    def __init__(self, controlDir):
        global vcs, annotateClass, basicDiffClasses
        vcs = vcsClass(controlDir)
        annotateClass = self.annotateClasses()[0]
        basicDiffClasses = [c for c in self.diffClasses() if not c.recursive]

    def getMenuNames(self):
        return [vcs.name]

    def getInteractiveActionClasses(self, dynamic):
        return [LogGUI, LogGUIRecursive] + self.diffClasses() + [StatusGUI, StatusGUIRecursive] +\
            self.annotateClasses() + [AddGUI, AddGUIRecursive, self.getUpdateClass()]

    def annotateClasses(self):
        return [AnnotateGUI, AnnotateGUIRecursive]

    def diffClasses(self):
        return [DiffGUI, DiffGUIRecursive, FilteredDiffGUI, FilteredDiffGUIRecursive]

    def getRenameTestClass(self):
        return RenameTest

    def getUpdateClass(self):
        return UpdateGUI

    def getDefaultAccelerators(self):
        dict = {}
        accelKey = vcs.name + "_" + self.getUpdateClass()._getTitle()
        dict[accelKey.lower()] = "<control>u"
        return dict

    def getReplacements(self):
        return {adminactions.RemoveTests: RemoveTests,
                adminactions.RemoveFiles: RemoveFiles,
                adminactions.RemoveTestsForPopup: RemoveTestsForPopup,
                adminactions.RenameTest: self.getRenameTestClass(),
                adminactions.RenameFile: RenameFile,
                adminactions.PasteTests: PasteTests,
                changeteststate.FindKnownBugs: FindKnownBugs}



"""
The default configuration's worth of action GUIs, implementing the interfaces in guiplugins
"""

try:
    # For backwards compatibility, don't require derived modules to know the internal structure here
    from .helpdialogs import *
    from .adminactions import *
    from .fileviewers import *
    from .fileproperties import *
    from .selectandfilter import *
    from .runningactions import *
    from .changeteststate import *
    from .housekeeping import *

    from ..guiplugins import InteractiveActionConfig as BaseInteractiveActionConfig
except ImportError as e:
    # Might want the default accelerators, don't crash if so
    if "No module named gtk" in str(e):
        class BaseInteractiveActionConfig:
            pass
    else:  # pragma: no cover - debugging aid only
        raise


class InteractiveActionConfig(BaseInteractiveActionConfig):
    def getMenuNames(self):
        return ["file", "edit", "view", "actions", "reorder", "help"]

    def getInteractiveActionClasses(self, dynamic):
        classes = housekeeping.getInteractiveActionClasses(dynamic)
        if dynamic:
            classes += changeteststate.getInteractiveActionClasses()
        else:
            classes += adminactions.getInteractiveActionClasses()

        classes += selectandfilter.getInteractiveActionClasses(dynamic)
        classes += runningactions.getInteractiveActionClasses(dynamic)
        classes += helpdialogs.getInteractiveActionClasses()
        classes += fileproperties.getInteractiveActionClasses()
        classes += fileviewers.getInteractiveActionClasses(dynamic)
        return classes

    def getDefaultAccelerators(self):
        dict = {}
        dict["quit"] = "<control>q"
        dict["select"] = "<control><alt>f"
        dict["filter"] = "<control><shift>f"
        dict["approve"] = "<control>s"
        dict["approve_as"] = "<control><alt>s"
        dict["copy"] = "<control>c"
        dict["kill"] = "<control>Delete"
        dict["remove_tests"] = "<control>Delete"
        dict["cut"] = "<control>x"
        dict["paste"] = "<control>v"
        dict["save_selection"] = "<control>d"
        dict["load_selection"] = "<control><shift>o"
        dict["reset"] = "<control>e"
        dict["reconnect"] = "<control><shift>r"
        dict["run"] = "<control>r"
        dict["rerun"] = "<control>r"
        dict["rename_test"] = "<control>m"
        dict["refresh"] = "F5"
        dict["record_use-case"] = "F9"
        dict["recompute_status"] = "F5"
        dict["add_test"] = "<control>n"
        dict["enter_failure_information"] = "<control>i"
        dict["move_down"] = "<control>Page_Down"
        dict["move_up"] = "<control>Page_Up"
        dict["move_to_first"] = "<control>Home"
        dict["move_to_last"] = "<control>End"
        dict["mark"] = "<control><shift>m"
        dict["unmark"] = "<control><shift>u"
        return dict




# One basic action that wants to bypass the GTK imports

from texttestlib.default.scripts import DocumentConfig


class DocumentGUIConfig(DocumentConfig):
    def setUpApplication(self, app):
        self.reloadForOverrideOs(app)
        from .guiutils import GUIConfig
        from .default_gui import InteractiveActionConfig
        guiConfig = GUIConfig(False, [app], GUIConfig.getDefaultColours(),
                              InteractiveActionConfig().getDefaultAccelerators())
        for key in sorted(guiConfig.configDir.keys()):
            docOutput = guiConfig.configDocs[key]
            value = guiConfig.configDir[key]
            print(key + "|" + self.getValue(value) + "|" + docOutput)

    def getValue(self, value):
        class ArgWrapper:
            def __init__(self, val):
                self.val = val

            def __repr__(self):
                return str(self.val)

        if isinstance(value, dict):
            newdict = {}
            # Avoid revolting floating point output in Python 2.6, this isn't
            # necessary in 2.7 and later
            for key, val in list(value.items()):
                if isinstance(val, float):
                    newdict[key] = ArgWrapper(val)
                else:
                    newdict[key] = val
            return str(newdict)
        else:
            return str(value)




""" All the standard scripts that come with the default configuration """

from . import sandbox
import operator
import os
import shutil
import sys
import random
from glob import glob
from texttestlib import plugins
from collections import OrderedDict
from configparser import RawConfigParser
from functools import reduce
from pprint import pformat


class CountTest(plugins.Action):
    scriptDoc = "report on the number of tests selected, by application"
    appCount = OrderedDict()

    @classmethod
    def finalise(self):
        for app, count in list(self.appCount.items()):
            print(app.description(), "has", count, "tests")
        print("There are", sum(self.appCount.values()), "tests in total.")

    def __repr__(self):
        return "Counting"

    def __call__(self, test):
        self.describe(test)
        self.appCount[test.app] += 1

    def setUpSuite(self, suite):
        self.describe(suite)

    def setUpApplication(self, app):
        self.appCount[app] = 0


class WriteDividedSelections(plugins.ScriptWithArgs):
    scriptDoc = "divide the test suite into equally sized selections, for parallel testing without communication possibilities"
    files = []
    counts = []

    def __init__(self, args=[]):
        if len(self.files) == 0:
            WriteDividedSelections.initialise(args)

    @classmethod
    def initialise(cls, args):
        argDict = cls.parseArguments(args, ["count", "prefix"])
        prefix = argDict["prefix"]
        for fn in glob(prefix + "_*"):
            os.remove(fn)
        for i in range(int(argDict["count"])):
            fn = prefix + "_" + str(i + 1)
            f = open(fn, "a")
            f.write("-tp ")
            cls.files.append(f)
        cls.counts = [0] * len(cls.files)

    def setUpSuite(self, suite):
        if suite.parent is None:
            for f in self.files:
                f.write("appdata=" + suite.app.name + suite.app.versionSuffix() + "\n")

    def __call__(self, test):
        minCount = min(self.counts)
        minIndices = [i for (i, count) in enumerate(self.counts) if count == minCount]
        chosenIndex = minIndices[int(random.random() * len(minIndices))] if len(minIndices) > 1 else minIndices[0]
        self.counts[chosenIndex] += 1
        self.files[chosenIndex].write(test.getRelPath() + "\n")


class DocumentOptions(plugins.Action):
    multiValueOptions = ["a", "c", "f", "funion", "fintersect", "t", "ts", "v"]

    def setUpApplication(self, app):
        groups = app.createOptionGroups([app])
        keys = reduce(operator.add, (list(g.keys()) for g in groups), [])
        keys.sort()
        for key in keys:
            self.displayKey(key, groups)

    def displayKey(self, key, groups):
        for group in groups:
            option = group.getOption(key)
            if option:
                keyOutput, docOutput = self.optionOutput(key, group, option)
                self.display(keyOutput, self.groupOutput(group), docOutput)

    def display(self, keyOutput, groupOutput, docOutput):
        if not docOutput.startswith("Private"):
            print(keyOutput + ";" + groupOutput + ";" + docOutput.replace("SGE", "SGE/LSF"))

    def groupOutput(self, group):
        knownGroups = ["Selection", "Basic", "Advanced"]
        if group.name == "Invisible":
            return "N/A"
        elif group.name in knownGroups:
            return group.name
        else:
            return "Advanced"

    def optionOutput(self, key, group, option):
        keyOutput = "-" + key
        docs = option.name
        if isinstance(option, plugins.TextOption):
            keyOutput += " <value>"
            if (docs == "Execution time"):
                keyOutput = "-" + key + " <time specification string>"
            else:
                docs += " <value>"
            if key in self.multiValueOptions:
                keyOutput += ",..."
                docs += ",..."

        if group.name.startswith("Select"):
            return keyOutput, "Select " + docs.lower()
        else:
            return keyOutput, docs


class DocumentConfig(plugins.ScriptWithArgs):
    def __init__(self, args=[]):
        argDict = self.parseArguments(args, ["os", "entries"])
        self.onlyEntries = []
        if "entries" in argDict:
            self.onlyEntries = argDict["entries"].split(",")
        self.overrideOs = argDict.get("os") if "os" in argDict else None
        
    def getEntriesToUse(self, app):
        if len(self.onlyEntries) > 0:
            return self.onlyEntries
        else:
            return sorted(list(app.configDir.keys()) + list(app.configDir.aliases.keys()))

    def reloadForOverrideOs(self, app):
        if self.overrideOs and self.overrideOs != os.name:
            realOs = self.overrideOs
            os.name = self.overrideOs
            app.reloadConfiguration()
            os.name = realOs

    def setUpApplication(self, app):
        self.reloadForOverrideOs(app)
        for key in self.getEntriesToUse(app):
            realKey = app.configDir.aliases.get(key, key)
            if realKey == key:
                docOutput = app.configDocs.get(realKey, "NO DOCS PROVIDED")
            else:
                docOutput = "Alias. See entry for '" + realKey + "'"
            if not docOutput.startswith("Private"):
                value = app.configDir[realKey]
                print(key + "|" + self.interpretArgument(value) + "|" + docOutput)

    def interpretArgument(self, arg):
        argStr = pformat(arg, width=1000) if isinstance(arg, dict) else str(arg)
        if os.sep == "\\":
            # in python strings get double backslashes, handle this
            doubleBackslashRoot = plugins.installationRoots[0].replace("\\", "\\\\")
            argStr = argStr.replace(doubleBackslashRoot, "<source library>")
        return argStr.replace(plugins.installationRoots[0], "<source library>")


class DocumentEnvironment(plugins.Action):
    def __init__(self, args=[]):
        self.onlyEntries = args
        self.prefixes = ["TEXTTEST_", "USECASE_", "STORYTEXT_"]
        self.exceptions = ["TEXTTEST_PERSONAL_", "STORYTEXT_HOME_LOCAL", "TEXTTEST_FAKE_USER"]

    def getEntriesToUse(self, app):
        rootDir = plugins.installationRoots[0]
        return self.findAllVariables(app, self.prefixes, rootDir)

    def findAllVariables(self, app, prefixes, rootDir):
        includeSite = app.inputOptions.configPathOptions()[0]
        allVars = {}
        for root, dirs, files in os.walk(rootDir):
            if "log" in dirs:
                dirs.remove("log")
            if "storytext" in dirs:
                dirs.remove("storytext")
            if not includeSite and "site" in dirs:
                dirs.remove("site")

            # We only want to document environment variables in the configuration we're using here, not every configuration we can find
            if root.endswith("lib"):
                toRemove = []
                for dir in dirs:
                    if dir != "libexec" and dir not in sys.modules and "texttestlib." + dir not in sys.modules:
                        toRemove.append(dir)
                for dir in toRemove:
                    dirs.remove(dir)
            for file in files:
                if file.endswith(".py"):
                    path = os.path.join(root, file)
                    self.findVarsInFile(path, allVars, prefixes)
        return allVars

    def getArgList(self, line, functionName):
        pos = line.find(functionName) + len(functionName)
        parts = line[pos:].strip().split("#")
        endPos = parts[0].find(")")
        argStr = parts[0][:endPos + 1]
        for _ in range(argStr.count("(", 1)):
            endPos = parts[0].find(")", endPos + 1)
            argStr = parts[0][:endPos + 1]
        allArgs = self.getActualArguments(argStr)
        if len(parts) > 1:
            allArgs.append(parts[1].strip())
        else:
            allArgs.append("")
        return allArgs

    def getActualArguments(self, argStr):
        if not argStr.startswith("("):
            return []

        # Pick up app.getConfigValue
        class FakeApp:
            def getConfigValue(self, name):
                return "Config value '" + name + "'"
        app = FakeApp()  # @UnusedVariable
        try:
            argTuple = eval(argStr)
            if type(argTuple) == tuple:
                allArgs = list(eval(argStr))
                return [ str(allArgs[1]) ]
            else:
                return []
        except Exception:  # could be anything at all
            if argStr.endswith("executable)"):
                return [ "<source directory>/bin/texttest" ]
            else:
                return []

    def isRelevant(self, var, vars, prefixes):
        if var in self.exceptions or var in prefixes or "SLEEP" in var or "SELFTEST" in var or \
                (len(self.onlyEntries) > 0 and var not in self.onlyEntries):
            return False
        prevVal = vars.get(var, [])
        return not prevVal or not prevVal[0]

    def findVarsInFile(self, path, vars, prefixes):
        import re
        regexes = [re.compile("([^/ \"'\.,\(]*)[\(]?[\"'](" + prefix + "[^/ \"'\.,]*)") for prefix in prefixes]
        for line in open(path):
            for regex in regexes:
                match = regex.search(line)
                if match is not None:
                    functionName = match.group(1)
                    var = match.group(2).strip()
                    if self.isRelevant(var, vars, prefixes):
                        argList = self.getArgList(line, functionName)
                        vars[var] = argList

    def setUpApplication(self, app):
        vars = self.getEntriesToUse(app)
        print("The following variables may be set by the user :")
        for key in sorted(vars.keys()):
            argList = vars[key]
            if len(argList) > 1:
                print(key + "|" + "|".join(argList))

        print("The following variables are set by TextTest :")
        for var in sorted([key for key in list(vars.keys()) if len(vars[key]) == 1]):
            print(var + "|" + vars[var][0])


class DocumentScripts(plugins.Action):
    def setUpApplication(self, app):
        from . import batch, comparetest, performance
        for module in [batch, comparetest, sys.modules[__name__], performance]:
            names = dir(module)
            for name in names:
                try:
                    docString = getattr(getattr(module, name), "scriptDoc")
                    scriptName = module.__name__.replace("texttestlib.default.", "").replace(
                        "scripts", "default") + "." + name
                    print(scriptName + "|" + docString)
                except AttributeError:
                    pass


class ReplaceText(plugins.ScriptWithArgs):
    scriptDoc = "Perform a search and replace on all files with the given stem"

    def __init__(self, args):
        argDict = self.parseArguments(args, ["old", "new", "file", "regexp", "argsReplacement", "includeShortcuts"])
        tryAsRegexp = "regexp" not in argDict or argDict["regexp"] == "1"
        self.argsReplacement = "argsReplacement" in argDict and argDict["argsReplacement"] == "1"
        self.oldText = argDict["old"].replace("\\n", "\n")
        self.newText = argDict["new"].replace("\\n", "\n")
        if self.newText.endswith("\n") and self.oldText.endswith("\n"):
            self.oldText = self.oldText.rstrip()
        self.newText = self.newText.rstrip()
        self.trigger = plugins.MultilineTextTrigger(self.oldText, tryAsRegexp, False) if not self.argsReplacement else plugins.TextTrigger(self.oldText, tryAsRegexp, False)
        self.newMultiLineText = self.newText.split("\n")
        self.stems = []
        fileStr = argDict.get("file")
        if fileStr:
            self.stems = plugins.commasplit(fileStr)
        self.includeShortcuts = "includeShortcuts" in argDict and argDict["includeShortcuts"] == "1"

    def __repr__(self):
        return "Replacing " + self.oldText + " with " + self.newText + " for"

    def getFilesToChange(self, test, stem):
        if self.includeShortcuts and test.app.name == "shortcut" and test.classId() == "test-case":
            return [test.dircache.pathName(f) for f in test.dircache.contents if f.endswith(".shortcut")]
        else:
            return test.getFileNamesMatching(stem)

    def __call__(self, test):
        self.replaceInFiles(test)

    def replaceInFiles(self, test):
        replaced = False
        for stem in self.stems:
            for stdFile in self.getFilesToChange(test, stem):
                if os.path.isfile(stdFile) and not plugins.containsAutoGeneratedText(stdFile):
                    self.replaceInFile(test, stdFile)
                    replaced = True
        return replaced

    def replaceInFile(self, test, stdFile):
        fileName = os.path.basename(stdFile)
        self.describe(test, " - file " + fileName)
        sys.stdout.flush()
        unversionedFileName = ".".join(fileName.split(".")[:2])
        if "." not in unversionedFileName:
            unversionedFileName += "." + test.app.name
        tmpFile = os.path.join(test.getDirectory(temporary=1), unversionedFileName)
        with open(tmpFile, "w") as writeFile:
            with open(stdFile) as readFile:
                for line in readFile:
                    newLine = self.trigger.replace(line, self.newMultiLineText if not self.argsReplacement else self.replaceArgs)
                    writeFile.write(newLine)
                if self.oldText[-1] != "\n" and not self.argsReplacement:
                    writeFile.write(self.trigger.getLeftoverText())

    def replaceArgs(self, matchobj):
        from storytext.replayer import ReplayScript
        return ReplayScript.getTextWithArgs(self.newText, [arg for arg in matchobj.groups()])

    def usesComparator(self):
        return True

    def comparesSuites(self, app):
        defStems = app.defFileStems()
        return any((stem in defStems for stem in self.stems))

    def setUpSuite(self, suite):
        if not self.replaceInFiles(suite):
            self.describe(suite)

    def setUpApplication(self, app):
        if len(self.stems) == 0:
            logFile = app.getConfigValue("log_file")
            if not logFile in self.stems:
                self.stems.append(logFile)


class ExportTests(plugins.ScriptWithArgs):
    scriptDoc = "Export the selected tests to a different test suite"

    def __init__(self, args):
        argDict = self.parseArguments(args, ["dest"])
        self.otherTTHome = argDict.get("dest")
        self.otherSuites = {}
        self.placements = {}
        if not self.otherTTHome:
            raise plugins.TextTestError("Must provide 'dest' argument to indicate where tests should be exported")

    def __repr__(self):
        return "Checking for export of"

    def __call__(self, test):
        self.tryExport(test)

    def setUpSuite(self, suite):
        self.placements[suite] = 0
        if suite.parent:
            self.tryExport(suite)

    def tryExport(self, test):
        otherRootSuite = self.otherSuites.get(test.app)
        otherTest = otherRootSuite.findSubtestWithPath(test.getRelPath())
        parent = test.parent
        if otherTest:
            self.describe(test, " - already exists")
        else:
            otherParent = otherRootSuite.findSubtestWithPath(parent.getRelPath())
            if otherParent:
                self.describe(test, " - CREATING...")
                self.copyTest(test, otherParent, self.placements[parent])
            else:
                self.describe(test, " - COULDN'T FIND PARENT")
        self.placements[parent] += 1

    def copyTest(self, test, otherParent, placement):
        # Do this first, so that if it fails due to e.g. full disk, we won't register the test either...
        testDir = otherParent.makeSubDirectory(test.name)
        self.copyTestContents(test, testDir)
        otherParent.registerTest(test.name, test.description, placement)
        otherParent.addTest(test.__class__, test.name, test.description, placement)

    def copyTestContents(self, test, newDir):
        stdFiles, defFiles = test.listApprovedFiles(allVersions=True)
        for sourceFile in stdFiles + defFiles:
            dirname, local = os.path.split(sourceFile)
            if dirname == test.getDirectory():
                targetFile = os.path.join(newDir, local)
                shutil.copy2(sourceFile, targetFile)

        extFiles = test.listExternallyEditedFiles()[1]
        dataFiles = test.listDataFiles() + extFiles
        for sourcePath in dataFiles:
            if os.path.isdir(sourcePath):
                continue
            targetPath = sourcePath.replace(test.getDirectory(), newDir)
            plugins.ensureDirExistsForFile(targetPath)
            shutil.copy2(sourcePath, targetPath)

    def setUpApplication(self, app):
        self.otherSuites[app] = app.createExtraTestSuite(otherDir=self.otherTTHome)

# A standalone action, we add description and generate the main file instead...


class ExtractStandardPerformance(sandbox.ExtractPerformanceFiles):
    scriptDoc = "update the standard performance files from the standard log files"

    def __init__(self):
        sandbox.ExtractPerformanceFiles.__init__(self, sandbox.MachineInfoFinder())

    def __repr__(self):
        return "Extracting standard performance for"

    def __call__(self, test):
        self.describe(test)
        sandbox.ExtractPerformanceFiles.__call__(self, test)

    def findLogFiles(self, test, stem):
        return test.getFileNamesMatching(stem)

    def getFileToWrite(self, test, stem):
        name = stem + "." + test.app.name + test.app.versionSuffix()
        return os.path.join(test.getDirectory(), name)

    def allMachinesTestPerformance(self, test, fileStem):
        # Assume this is OK: the current host is in any case utterly irrelevant
        return 1

    def setUpSuite(self, suite):
        self.describe(suite)

    def getMachineContents(self, test):
        return " on unknown machine (extracted)\n"


class InsertShortcuts(plugins.ScriptWithArgs):
    def __repr__(self):
        return "Inserting shortcuts into usecases for"

    def __call__(self, test):
        stdFiles = test.getFileNamesMatching("*usecase*")
        for stdFile in stdFiles:
            fileName = os.path.basename(stdFile)
            self.describe(test, " - file " + fileName)
            unversionedFileName = ".".join(fileName.split(".")[:2])
            tmpFile = os.path.join(test.getDirectory(temporary=1), unversionedFileName)
            storytextHome = test.getEnvironment("STORYTEXT_HOME")
            recordScript = self.getRecordScript(tmpFile, storytextHome)
            with open(stdFile, "rU") as readFile:
                for line in readFile:
                    recordScript.record(line.strip("\n"))

    def getRecordScript(self, stdFile, storytextHome):
        from storytext.replayer import ShortcutManager
        from storytext import scriptEngine
        from storytext.recorder import RecordScript
        return RecordScript(stdFile, scriptEngine.getShortcuts(storytextHome))

    def usesComparator(self):
        return True

    def comparesSuites(self, *args):
        return False


class FilterUIMapFile(plugins.ScriptWithArgs):
    scriptDoc = "Edit the UI map to remove all unused entries, list the unused shortcuts (GUI testing)"
    instancesByFile = OrderedDict()

    def __init__(self):
        self.uiMapFileHandler = None
        self.uiMapFileUsed = {}
        from storytext.replayer import ShortcutManager
        self.shortcutManager = ShortcutManager()
        self.shortcutsUsed = set()

    def __repr__(self):
        return "Filtering UI map file for"

    def __call__(self, test):
        storytextHome = test.getEnvironment("STORYTEXT_HOME")
        uiMapFile = os.path.join(storytextHome, "ui_map.conf")
        self.instancesByFile[uiMapFile].filterUseCaseCommands(test)

    def storeUsage(self, script):
        for command in script.commands:
            shortcut, _ = self.shortcutManager.findShortcut(command)
            if shortcut:
                self.shortcutsUsed.add(shortcut)
            else:
                for section, option in self.uiMapFileHandler.findSectionsAndOptions(command):
                    self.uiMapFileUsed.setdefault(section, []).append(option)

    def filterUseCaseCommands(self, test):
        stdFiles = test.getFileNamesMatching("*usecase*")
        from storytext.replayer import ReplayScript
        for stdFile in sorted(stdFiles):
            fileName = os.path.basename(stdFile)
            self.describe(test, " - file " + fileName)
            script = ReplayScript(stdFile, ignoreComments=True)
            self.storeUsage(script)

    def setUpSuite(self, suite):
        if suite.parent is None:
            storytextHome = suite.getEnvironment("STORYTEXT_HOME")
            uiMapFile = os.path.join(storytextHome, "ui_map.conf")
            if not uiMapFile in self.instancesByFile:
                self.instancesByFile[uiMapFile] = self
                from storytext.guishared import UIMapFileHandler
                from storytext.scriptengine import ScriptEngine
                self.uiMapFileHandler = UIMapFileHandler([uiMapFile])
                for shortcut in ScriptEngine.getShortcuts(storytextHome):
                    self.shortcutManager.add(shortcut)
        else:
            self.describe(suite)

    @classmethod
    def finalise(cls):
        for uiMapFile, instance in list(cls.instancesByFile.items()):
            instance.writeResults(uiMapFile)

    def writeResults(self, uiMapFile):
        for name, shortcut in self.shortcutManager.getShortcuts():
            if shortcut in self.shortcutsUsed:
                self.storeUsage(shortcut)
            else:
                print("Shortcut not used", shortcut.name)

        writeParser = RawConfigParser(dict_type=OrderedDict)
        writeParser.optionxform = str
        for section in self.uiMapFileHandler.sections():
            usedActions = self.uiMapFileUsed.get(section, [])
            entriesToAdd = [action_event for action_event in self.uiMapFileHandler.items(
                section) if action_event[0] in usedActions or len(action_event[1].strip()) == 0]
            if entriesToAdd:
                writeParser.add_section(section)
                for actionName, eventName in entriesToAdd:
                    writeParser.set(section, actionName, eventName)
        print("Updating UI map file...")
        writeParser.write(open(uiMapFile, "w"))
